"""an executable to calculate p(astro) values and upload to GraceDB events"""

# Copyright (C)           Becca Ewing
# Copyright (C) 2024-2025 Yun-Jing Huang

__usage__ = "sgnl-ll-inspiral-pastro-uploader [--options]"
__author__ = (
    "Rebecca Ewing (rebecca.ewing@ligo.org), "
    "Yun-Jing Huang (yun-jing.huang@ligo.org)"
)

# -------------------------------------------------
#                   Preamble
# -------------------------------------------------

import copy
import http.client as httplib
import json
import logging
import os
import shutil
import tempfile
import time
from argparse import ArgumentParser
from collections import deque
from io import BytesIO

from igwn_ligolw import ligolw, lsctables
from igwn_ligolw import utils as ligolw_utils
from igwn_ligolw.array import use_in as array_use_in
from igwn_ligolw.param import use_in as param_use_in
from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.scald import utils
from pastro import pastro  # FIXME move pastro to sgn?

from sgnl import events
from sgnl.gracedb import FakeGracedbClient


@lsctables.use_in
@array_use_in
@param_use_in
class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
    pass


lsctables.use_in(LIGOLWContentHandler)

# -------------------------------------------------
#                   Functions
# -------------------------------------------------


def parse_command_line():

    parser = ArgumentParser(usage=__usage__, description=__doc__)
    parser.add_argument(
        "-v", "--verbose", default=False, action="store_true", help="Be verbose."
    )
    parser.add_argument(
        "--num-messages",
        type=int,
        default=10000,
        help="number of messages to process per cadence",
    )
    parser.add_argument(
        "--tag",
        metavar="string",
        default="test",
        help="Sets the name of the tag used. Default = 'test'",
    )
    parser.add_argument(
        "--processing-cadence",
        type=float,
        default=0.1,
        help="Rate at which the event uploader acquires and processes data. Default = "
        "0.1 seconds.",
    )
    parser.add_argument(
        "--request-timeout",
        type=float,
        default=0.2,
        help="Timeout for requesting messages from a topic. Default = 0.2 seconds.",
    )
    parser.add_argument(
        "--kafka-server",
        metavar="string",
        help="Sets the server url that the kafka topic is hosted on. Required.",
    )
    parser.add_argument(
        "--input-topic",
        metavar="string",
        action="append",
        help="Sets the input kafka topic(s). Required.",
    )
    parser.add_argument(
        "--gracedb-service-url",
        metavar="url",
        help="Override default GracedB service url.",
    )
    parser.add_argument(
        "--pastro-filename",
        metavar="string",
        default="p_astro.json",
        help="Name to upload the p(astro) file with. Default is p_astro.json.",
    )
    parser.add_argument(
        "--model-name", metavar="string", help="Name of pastro model used, eg FGMC."
    )
    parser.add_argument(
        "--pastro-model-file", metavar="file", help="Filename of pastro model."
    )
    parser.add_argument(
        "--rank-stat",
        metavar="file",
        help="Filename of ranking stat pdf to update the pastro model with.",
    )

    options = parser.parse_args()

    return options


# -------------------------------------------------
#                    Classes
# -------------------------------------------------


class PAstroUploader(events.EventProcessor):
    def __init__(
        self,
        gracedb_service_url: str,
        input_topic: list[str],
        kafka_server: str,
        logger: logging.Logger,
        model_name: str,
        pastro_model_file: str,
        rank_stat: str,
        num_messages: int = 10000,
        pastro_filename: str = "p_astro.json",
        processing_cadence: float = 0.1,
        request_timeout: float = 0.2,
        tag: str = "test",
    ):
        self.logger = logger
        self.logger.info("setting up pastro uploader...")

        self.tag = tag
        self.model_name = model_name
        self.max_retries = 3
        self.filename = pastro_filename
        self.pastro_model_file = pastro_model_file
        self.p_astro_topic = f"sgnl.{self.tag}.p_astro"

        self.rank_stat = rank_stat
        self.last_rankstat_update = None
        self.update_rankstat_cadence = 4.0 * 3600.0

        self.is_injection_job = input_topic[0] == "inj_uploads"
        heartbeat_topic = (
            f"sgnl.{tag}.pastro_uploader_heartbeat"
            if not self.is_injection_job
            else f"sgnl.{tag}.inj_pastro_uploader_heartbeat"
        )

        # init the EventProcessor
        events.EventProcessor.__init__(
            self,
            process_cadence=processing_cadence,
            request_timeout=request_timeout,
            num_messages=num_messages,
            kafka_server=kafka_server,
            input_topic=[f"sgnl.{tag}.{topic}" for topic in input_topic],
            tag="-".join([self.tag, self.model_name]),
            send_heartbeats=True,
            heartbeat_cadence=60.0,
            heartbeat_topic=heartbeat_topic,
        )

        # set up gracedb client
        if gracedb_service_url.startswith("file"):
            self.client = FakeGracedbClient(gracedb_service_url)
        else:
            self.client = GraceDb(gracedb_service_url)

        # load model file
        self.model = self.load_model(self.pastro_model_file)
        # attempt to initialize model with ranking stat info
        self.add_rankstat_to_model()

        # start a list of events
        self.events: deque = deque(maxlen=100)

    def ingest(self, message):
        # load the message value
        msg = json.loads(message.value())

        # append to list of messages to be handled
        # these messages come post aggregation, so
        # we will calculate a pastro for each one
        # without checking for duplicates
        self.events.append(
            {
                "gid": msg["gid"],
                "time": msg["time"] + msg["time_ns"] * 10.0**-9.0,
                "coinc": self.load_xmlobj(msg["coinc"]),
                "pastro": None,
                "snr_optimized": msg["snr_optimized"],
                "upload_attempts": 0,
                "upload_success": False,
            }
        )

        # update pastro model with ranking stat info
        t = utils.gps_now()
        if (
            not self.last_rankstat_update
            or t - self.last_rankstat_update >= self.update_rankstat_cadence
        ):
            self.add_rankstat_to_model()

    def handle(self):
        # if we have not updated the model with
        # ranking stat info yet, we cant
        # accurately compute pastros. Check that
        # this is done before continuing.
        if not self.last_rankstat_update:
            return

        # for all events in the list, calculate
        # the pastro and upload to GraceDB
        for event in copy.copy(self.events):
            if (
                not event["upload_success"]
                and event["upload_attempts"] < self.max_retries
            ):
                self.logger.debug("Processing %s...", event["gid"])
                pastro = self.calculate_pastro(event)
                # FIXME: hard code this value until pastro dependency is resolved
                event["pastro"] = pastro

                # attempt to upload the file
                response = self.upload_file(
                    event["gid"],
                    "SGNL internally computed p-astro",
                    self.filename,
                    pastro,
                    f"{self.model_name}_p_astro",
                )

                # after successful upload, add label
                if response:
                    try:
                        response = self.client.writeLabel(event["gid"], "PASTRO_READY")
                    except HTTPError:
                        self.logger.exception("except HTTPError")
                    else:
                        # remove this event from the list
                        self.logger.debug(
                            "Successfully uploaded %s to %s.",
                            self.filename,
                            event["gid"],
                        )
                        event["upload_success"] = True
                        self.send_p_astro(event, pastro)
                else:
                    self.logger.debug(
                        "Failed to upload %s to %s.", self.filename, event["gid"]
                    )
                    event["upload_attempts"] += 1

    def load_xmlobj(self, xmlobj):
        if isinstance(xmlobj, str):
            xmlobj = BytesIO(xmlobj.encode("utf-8"))
        return ligolw_utils.load_fileobj(xmlobj, contenthandler=LIGOLWContentHandler)

    def load_model(self, filename):
        model = pastro.load(filename)
        model.finalize(model.prior())
        return model

    def add_rankstat_to_model(self):
        # update the pastro model with ranking stat information
        # from the current analysis in the path provided by
        # self.rank_stat. Note: this will fail if the analysis
        # hasnt burned in yet, since there wont be a ranking stat
        # pdf file yet. Hence the try/except with OSError
        try:
            self.model.update_rankstatpdf(self.rank_stat)
            self.model.finalize(self.model.prior())

            tmp, tmp_path = tempfile.mkstemp(
                ".h5", dir=os.getenv("_CONDOR_SCRATCH_DIR", tempfile.gettempdir())
            )
            os.close(tmp)

            self.logger.debug("Writing model to %s...", tmp_path)
            self.model.to_h5(tmp_path)

            self.logger.debug("Moving %s to %s...", tmp_path, self.pastro_model_file)
            shutil.move(tmp_path, self.pastro_model_file)

            self.last_rankstat_update = utils.gps_now()

        except (OSError, ValueError):
            self.logger.exception("Failed to update model with ranking stat info")

        return

    def calculate_pastro(self, event):
        pa = None
        if not event["snr_optimized"]:
            data = self.parse_data_from_coinc(event["coinc"])
            if data:
                pa = self.model(data, inj=self.is_injection_job)
        # for snr optimized coincs from manifold, just copy the pastro
        # from the corresponding non-optimized gstlal coinc
        else:
            this_time = event["time"]
            for e in self.events:
                if e["time"] == this_time and not e["snr_optimized"]:
                    pa = e["pastro"]
        return pa

    def parse_data_from_coinc(self, coinc):
        try:
            mchirp = lsctables.CoincInspiralTable.get_table(coinc).getColumnByName(
                "mchirp"
            )[0]
            snr = lsctables.CoincInspiralTable.get_table(coinc).getColumnByName("snr")[
                0
            ]
            likelihood = lsctables.CoincTable.get_table(coinc).getColumnByName(
                "likelihood"
            )[0]
            template_id = lsctables.SnglInspiralTable.get_table(coinc).getColumnByName(
                "Gamma0"
            )[0]
            fhigh = lsctables.SnglInspiralTable.get_table(coinc).getColumnByName(
                "f_final"
            )[0]
        except Exception:
            self.logger.exception("Failed to parse coinc file.")

        else:
            return {
                "mchirp": mchirp,
                "likelihood": likelihood,
                "template_id": template_id,
                "snr": snr,
                "fhigh": fhigh,
            }

    def upload_file(self, graceid, message, filename, contents, tag, retries=3):
        for attempt in range(retries):
            try:
                # try to upload the file
                response = self.client.writeLog(
                    graceid,
                    message,
                    filename=filename,
                    filecontents=contents,
                    tagname=tag,
                )
            except HTTPError:
                # if the upload fails with an HTTPError, print it out
                self.logger.exception("except HTTPError")
            else:
                # if the upload does not fail check that the
                # response is what we want and leave this function
                if response.status == httplib.CREATED:
                    return response

            # if this upload attempt failed print a message, sleep for
            # 1 second before trying again
            self.logger.info(
                "upload of %s for %s failed on attempt %d / %d",
                filename,
                graceid,
                attempt,
                retries,
            )
            time.sleep(1.0)

        else:
            # if all uploads failed, return False
            self.logger.warning("upload of %s for %s failed.", filename, graceid)
            return False

    def send_p_astro(self, event, pastro):
        """
        send p(astro) via Kafka
        """
        p_astro = {"time": event["time"], "p_astro": pastro}
        self.producer.produce(topic=self.p_astro_topic, value=json.dumps(p_astro))
        # FIXME: the above is copied from gstlal, but should it actually be p_astro?
        self.producer.poll(0)


# -------------------------------------------------
#                     Main
# -------------------------------------------------


def main():
    # parse arguments
    options = parse_command_line()

    # set up logging
    log_level = logging.DEBUG if options.verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s | ll_inspiral_pastro_uploader : %(levelname)s : %(message)s"
    )
    logger = logging.getLogger("ll_inspiral_pastro_uploader")
    logger.setLevel(log_level)

    # create event uploader instance
    pastro_uploader = PAstroUploader(
        gracedb_service_url=options.gracedb_service_url,
        input_topic=options.input_topic,
        kafka_server=options.kafka_server,
        logger=logger,
        model_name=options.model_name,
        pastro_model_file=options.pastro_model_file,
        rank_stat=options.rank_stat,
        num_messages=options.num_messages,
        pastro_filename=options.pastro_filename,
        processing_cadence=options.processing_cadence,
        request_timeout=options.request_timeout,
        tag=options.tag,
    )

    # start up
    pastro_uploader.start()



"""Tests for sgnl.sinks.gracedb_sink module."""

from unittest import mock

import lal
import numpy
import pytest
from igwn_ligolw import ligolw, lsctables
from lal import LIGOTimeGPS

from sgnl.sinks import gracedb_sink


class TestCeilPow2:
    """Tests for ceil_pow_2 function."""

    def test_power_of_2_unchanged(self):
        """Test that powers of 2 are unchanged."""
        assert gracedb_sink.ceil_pow_2(2) == 2
        assert gracedb_sink.ceil_pow_2(4) == 4
        assert gracedb_sink.ceil_pow_2(8) == 8
        assert gracedb_sink.ceil_pow_2(16) == 16

    def test_round_up_to_power_of_2(self):
        """Test that values are rounded up to nearest power of 2."""
        assert gracedb_sink.ceil_pow_2(3) == 4
        assert gracedb_sink.ceil_pow_2(5) == 8
        assert gracedb_sink.ceil_pow_2(9) == 16
        assert gracedb_sink.ceil_pow_2(100) == 128

    def test_ceil_pow_2_one(self):
        """Test with value 1."""
        assert gracedb_sink.ceil_pow_2(1) == 1

    def test_ceil_pow_2_float(self):
        """Test with float values."""
        assert gracedb_sink.ceil_pow_2(3.5) == 4
        assert gracedb_sink.ceil_pow_2(100.1) == 128


class TestNsToGps:
    """Tests for ns_to_gps function."""

    def test_basic_conversion(self):
        """Test basic nanosecond to GPS conversion."""
        result = gracedb_sink.ns_to_gps(1000000000_000000000)
        assert isinstance(result, LIGOTimeGPS)
        assert int(result) == 1000000000

    def test_with_nanoseconds(self):
        """Test conversion with nanosecond component."""
        result = gracedb_sink.ns_to_gps(1000000000_500000000)
        assert int(result) == 1000000000
        assert result.gpsNanoSeconds == 500000000


class TestColMap:
    """Tests for col_map function."""

    def test_direct_mapping(self):
        """Test direct column mapping without renaming."""

        class MockRow:
            pass

        row = MockRow()
        datadict = {"col1": 100, "col2": "test"}
        gracedb_sink.col_map(row, datadict, {}, {})
        assert row.col1 == 100
        assert row.col2 == "test"

    def test_with_rename_mapping(self):
        """Test column mapping with rename."""

        class MockRow:
            pass

        row = MockRow()
        datadict = {"old_name": 42}
        mapdict = {"old_name": "new_name"}
        gracedb_sink.col_map(row, datadict, mapdict, {})
        assert row.new_name == 42
        assert not hasattr(row, "old_name")

    def test_exclude_column(self):
        """Test that columns mapped to None are excluded."""

        class MockRow:
            pass

        row = MockRow()
        datadict = {"keep_col": 1, "skip_col": 2}
        mapdict = {"skip_col": None}
        gracedb_sink.col_map(row, datadict, mapdict, {})
        assert row.keep_col == 1
        assert not hasattr(row, "skip_col")

    def test_with_function_mapping(self):
        """Test column mapping with function transform."""

        class MockRow:
            pass

        row = MockRow()
        datadict = {"value": 10}
        funcdict = {"value": lambda x: x * 2}
        gracedb_sink.col_map(row, datadict, {}, funcdict)
        assert row.value == 20


class TestAddTableWithNRows:
    """Tests for add_table_with_n_rows function."""

    def test_add_empty_table(self):
        """Test adding a table with 0 rows."""
        xmldoc = ligolw.Document()
        xmldoc.appendChild(ligolw.LIGO_LW())
        table = gracedb_sink.add_table_with_n_rows(xmldoc, lsctables.ProcessTable, 0)
        assert len(table) == 0

    def test_add_table_with_one_row(self):
        """Test adding a table with 1 row."""
        xmldoc = ligolw.Document()
        xmldoc.appendChild(ligolw.LIGO_LW())
        table = gracedb_sink.add_table_with_n_rows(xmldoc, lsctables.ProcessTable, 1)
        assert len(table) == 1

    def test_add_table_with_multiple_rows(self):
        """Test adding a table with multiple rows."""
        xmldoc = ligolw.Document()
        xmldoc.appendChild(ligolw.LIGO_LW())
        table = gracedb_sink.add_table_with_n_rows(xmldoc, lsctables.TimeSlideTable, 3)
        assert len(table) == 3


class TestLIGOLWContentHandler:
    """Tests for LIGOLWContentHandler class."""

    def test_content_handler_exists(self):
        """Test that LIGOLWContentHandler is properly defined."""
        assert gracedb_sink.LIGOLWContentHandler is not None
        assert issubclass(
            gracedb_sink.LIGOLWContentHandler, ligolw.LIGOLWContentHandler
        )


class TestBestEvent:
    """Tests for best_event function."""

    def test_empty_events(self):
        """Test with empty events list."""
        result = gracedb_sink.best_event([], [], [], 1.0, 1e-6)
        assert result == (None, None, None)

    def test_no_events_below_threshold(self):
        """Test when no events are below threshold."""
        events = [
            {"combined_far": 10.0, "network_snr": 8.0},
            {"combined_far": 20.0, "network_snr": 9.0},
        ]
        triggers = [[{"ifo": "H1"}], [{"ifo": "L1"}]]
        snr_ts = [None, None]
        result = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-6)
        assert result == (None, None, None)

    def test_event_below_opa_threshold(self):
        """Test selecting event below OPA threshold by max SNR."""
        events = [
            {"combined_far": 1e-7, "network_snr": 8.0},
            {"combined_far": 1e-8, "network_snr": 12.0},  # Higher SNR
        ]
        triggers = [[{"ifo": "H1"}], [{"ifo": "L1"}]]
        snr_ts = ["ts1", "ts2"]
        event, trigs, snr = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-6)
        assert event["network_snr"] == 12.0

    def test_event_below_regular_threshold(self):
        """Test selecting event below regular threshold by min FAR."""
        events = [
            {"combined_far": 0.5, "network_snr": 8.0},
            {"combined_far": 0.1, "network_snr": 6.0},  # Lower FAR
        ]
        triggers = [[{"ifo": "H1"}], [{"ifo": "L1"}]]
        snr_ts = ["ts1", "ts2"]
        event, trigs, snr = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-3)
        assert event["combined_far"] == 0.1

    def test_filters_none_combined_far(self):
        """Test that events with None combined_far are filtered out."""
        events = [
            {"combined_far": None, "network_snr": 100.0},
            {"combined_far": 0.5, "network_snr": 8.0},
        ]
        triggers = [[{"ifo": "H1"}], [{"ifo": "L1"}]]
        snr_ts = ["ts1", "ts2"]
        event, trigs, snr = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-3)
        assert event["combined_far"] == 0.5

    def test_filters_zero_combined_far(self):
        """Test that events with 0 combined_far are filtered out."""
        events = [
            {"combined_far": 0, "network_snr": 100.0},
            {"combined_far": 0.5, "network_snr": 8.0},
        ]
        triggers = [[{"ifo": "H1"}], [{"ifo": "L1"}]]
        snr_ts = ["ts1", "ts2"]
        event, trigs, snr = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-3)
        assert event["combined_far"] == 0.5

    def test_filters_none_triggers(self):
        """Test that None triggers are filtered from result."""
        events = [{"combined_far": 0.1, "network_snr": 8.0}]
        triggers = [[{"ifo": "H1"}, None, {"ifo": "L1"}]]
        snr_ts = ["ts1"]
        event, trigs, snr = gracedb_sink.best_event(events, triggers, snr_ts, 1.0, 1e-3)
        assert len(trigs) == 2


class TestSnglMap:
    """Tests for sngl_map function."""

    def test_basic_mapping(self):
        """Test basic sngl_map functionality."""

        class MockRow:
            pass

        class MockSngl:
            pass

        mock_sngl = MockSngl()
        for col in lsctables.SnglInspiralTable.validcolumns:
            col_name = col.split(":")[-1]
            setattr(mock_sngl, col_name, 0)
        mock_sngl.mass1 = 1.4
        mock_sngl.mass2 = 1.4

        sngls_dict = {1: mock_sngl}
        row = MockRow()

        gracedb_sink.sngl_map(row, sngls_dict, 1, exclude=())
        assert row.mass1 == 1.4
        assert row.mass2 == 1.4

    def test_mapping_with_exclude(self):
        """Test sngl_map with excluded columns."""

        class MockRow:
            pass

        class MockSngl:
            pass

        mock_sngl = MockSngl()
        for col in lsctables.SnglInspiralTable.validcolumns:
            col_name = col.split(":")[-1]
            setattr(mock_sngl, col_name, 0)
        mock_sngl.mass1 = 1.4

        sngls_dict = {1: mock_sngl}
        row = MockRow()

        gracedb_sink.sngl_map(row, sngls_dict, 1, exclude=("mass1",))
        assert not hasattr(row, "mass1")


class TestDownsampleSnr:
    """Tests for downsample_snr function."""

    def test_no_downsampling_needed(self):
        """Test when no downsampling is needed."""
        snr_ts = lal.CreateCOMPLEX8TimeSeries(
            name="snr",
            epoch=LIGOTimeGPS(1000000000),
            f0=0.0,
            deltaT=1.0 / 256,  # 256 Hz
            sampleUnits=lal.DimensionlessUnit,
            length=101,
        )
        snr_ts.data.data = numpy.ones(101, dtype=numpy.complex64)

        class MockSngl:
            f_final = 100.0  # ceil_pow_2(2*100) = 256

        result = gracedb_sink.downsample_snr(snr_ts, [MockSngl()])
        # Should return the same series since 256 == 256
        assert result is snr_ts

    def test_downsampling(self):
        """Test downsampling of SNR time series."""
        snr_ts = lal.CreateCOMPLEX8TimeSeries(
            name="snr",
            epoch=LIGOTimeGPS(1000000000),
            f0=0.0,
            deltaT=1.0 / 512,  # 512 Hz
            sampleUnits=lal.DimensionlessUnit,
            length=101,
        )
        snr_ts.data.data = numpy.arange(101, dtype=numpy.complex64)

        class MockSngl:
            f_final = 100.0  # ceil_pow_2(2*100) = 256

        result = gracedb_sink.downsample_snr(snr_ts, [MockSngl()])
        # Should downsample from 512 to 256 (factor of 2)
        assert result.deltaT == 1.0 / 256
        assert len(result.data.data) % 2 == 1


class TestWriteRankingstatXmldocGracedb:
    """Tests for write_rankingstat_xmldoc_gracedb function."""

    def test_empty_horizon_history(self, tmp_path):
        """Test with empty horizon history."""
        mock_rankingstat = mock.MagicMock()
        mock_rankingstat.copy.return_value = mock_rankingstat
        mock_rankingstat.terms = {"P_of_tref_Dh": mock.MagicMock()}
        mock_rankingstat.terms["P_of_tref_Dh"].horizon_history.maxkey.side_effect = (
            ValueError
        )

        output_file = str(tmp_path / "ranking.xml")
        gracedb_sink.write_rankingstat_xmldoc_gracedb(mock_rankingstat, output_file)

        mock_rankingstat.save.assert_called_once_with(output_file)

    def test_with_horizon_history(self, tmp_path):
        """Test with horizon history data."""
        mock_rankingstat = mock.MagicMock()
        mock_rankingstat.copy.return_value = mock_rankingstat
        mock_history = mock.MagicMock()
        mock_rankingstat.terms = {"P_of_tref_Dh": mock.MagicMock()}
        mock_rankingstat.terms["P_of_tref_Dh"].horizon_history.maxkey.return_value = (
            10000.0
        )
        mock_rankingstat.terms["P_of_tref_Dh"].horizon_history.values.return_value = [
            mock_history
        ]

        output_file = str(tmp_path / "ranking.xml")
        gracedb_sink.write_rankingstat_xmldoc_gracedb(mock_rankingstat, output_file)

        # Should trim history and save
        mock_history.__delitem__.assert_called()
        mock_rankingstat.save.assert_called_once_with(output_file)


class TestPublishGracedb:
    """Tests for publish_gracedb function."""

    @mock.patch("sgnl.sinks.gracedb_sink.ligolw_utils.write_fileobj")
    @mock.patch("sgnl.sinks.gracedb_sink.logo_data")
    def test_publish_gracedb(self, mock_logo, mock_write):
        """Test publishing to GraceDB."""
        mock_client = mock.MagicMock()
        mock_client.createEvent.return_value.json.return_value = {"graceid": "G123456"}
        mock_client.writeLog.return_value = {"status": "ok"}
        mock_logo.return_value = "dGVzdA=="  # base64 encoded "test"

        xmldoc = mock.MagicMock()

        gracedb_sink.publish_gracedb(
            mock_client,
            xmldoc,
            group="Test",
            pipeline="SGNL",
            search="AllSky",
            labels=["TEST"],
        )

        mock_client.createEvent.assert_called_once()
        mock_client.writeLog.assert_called_once()


class TestPublishKafka:
    """Tests for publish_kafka function."""

    @mock.patch("sgnl.sinks.gracedb_sink.ligolw_utils.write_fileobj")
    @mock.patch("sgnl.sinks.gracedb_sink.ligolw_utils.write_filename")
    @mock.patch("sgnl.sinks.gracedb_sink.write_rankingstat_xmldoc_gracedb")
    @mock.patch("os.makedirs")
    @mock.patch("os.path.exists")
    def test_publish_kafka_noninj(
        self,
        mock_exists,
        mock_makedirs,
        mock_write_ranking,
        mock_write_file,
        mock_write_obj,
    ):
        """Test publishing to Kafka for non-injection job."""
        mock_exists.return_value = False
        mock_client = mock.MagicMock()
        mock_strike_object = mock.MagicMock()
        mock_strike_object.likelihood_ratio_uploads = {"0001": mock.MagicMock()}

        # Create mock coinc_inspiral row
        mock_coinc_inspiral = mock.MagicMock()
        mock_coinc_inspiral.combined_far = 1e-10
        mock_coinc_inspiral.snr = 10.0
        mock_coinc_inspiral.end_time = 1000000000
        mock_coinc_inspiral.end_time_ns = 0
        mock_coinc_inspiral.mass = 2.8
        mock_coinc_inspiral.end = 1000000000.0

        # Create mock sngl_inspiral row
        mock_sngl = mock.MagicMock()
        mock_sngl.Gamma1 = 1.0

        # Mock the xmldoc and tables
        mock_xmldoc = mock.MagicMock()
        mock_coinc_table = mock.MagicMock()
        mock_coinc_table.__getitem__ = mock.MagicMock(return_value=mock_coinc_inspiral)
        mock_sngl_table = mock.MagicMock()
        mock_sngl_table.__getitem__ = mock.MagicMock(return_value=mock_sngl)

        with (
            mock.patch.object(
                lsctables.CoincInspiralTable, "get_table", return_value=mock_coinc_table
            ),
            mock.patch.object(
                lsctables.SnglInspiralTable, "get_table", return_value=mock_sngl_table
            ),
        ):
            gracedb_sink.publish_kafka(
                mock_client,
                mock_xmldoc,
                job_type="noninj_test",
                analysis_tag="test",
                instruments="H1L1",
                group="Test",
                search="AllSky",
                strike_object=mock_strike_object,
            )

        assert mock_client.produce.call_count == 2
        assert mock_client.poll.call_count == 2

    @mock.patch("sgnl.sinks.gracedb_sink.ligolw_utils.write_fileobj")
    @mock.patch("sgnl.sinks.gracedb_sink.ligolw_utils.write_filename")
    @mock.patch("sgnl.sinks.gracedb_sink.write_rankingstat_xmldoc_gracedb")
    @mock.patch("os.makedirs")
    @mock.patch("os.path.exists")
    def test_publish_kafka_inj(
        self,
        mock_exists,
        mock_makedirs,
        mock_write_ranking,
        mock_write_file,
        mock_write_obj,
    ):
        """Test publishing to Kafka for injection job."""
        mock_exists.return_value = True  # Directory already exists
        mock_client = mock.MagicMock()
        mock_strike_object = mock.MagicMock()
        mock_strike_object.likelihood_ratio_uploads = {"0001": mock.MagicMock()}

        mock_coinc_inspiral = mock.MagicMock()
        mock_coinc_inspiral.combined_far = 1e-10
        mock_coinc_inspiral.snr = 10.0
        mock_coinc_inspiral.end_time = 1000000000
        mock_coinc_inspiral.end_time_ns = 0
        mock_coinc_inspiral.mass = 2.8
        mock_coinc_inspiral.end = 1000000000.0

        mock_sngl = mock.MagicMock()
        mock_sngl.Gamma1 = 1.0

        mock_xmldoc = mock.MagicMock()
        mock_coinc_table = mock.MagicMock()
        mock_coinc_table.__getitem__ = mock.MagicMock(return_value=mock_coinc_inspiral)
        mock_sngl_table = mock.MagicMock()
        mock_sngl_table.__getitem__ = mock.MagicMock(return_value=mock_sngl)

        with (
            mock.patch.object(
                lsctables.CoincInspiralTable, "get_table", return_value=mock_coinc_table
            ),
            mock.patch.object(
                lsctables.SnglInspiralTable, "get_table", return_value=mock_sngl_table
            ),
        ):
            gracedb_sink.publish_kafka(
                mock_client,
                mock_xmldoc,
                job_type="inj_test",
                analysis_tag="test",
                instruments="H1L1",
                group="Test",
                search="AllSky",
                strike_object=mock_strike_object,
            )

        # Check that inj prefix is used
        call_args = mock_client.produce.call_args_list[0]
        assert "inj_events" in call_args[1]["topic"]
        # makedirs should not be called since dir exists
        mock_makedirs.assert_not_called()


class TestGraceDBSink:
    """Tests for GraceDBSink class."""

    @mock.patch("sgnl.sinks.gracedb_sink.Producer")
    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.__post_init__")
    def test_init_with_kafka(self, mock_super_init, mock_producer):
        """Test initialization with Kafka client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.strike_object = mock.MagicMock()
        sink.event_pad = "events"
        sink.spectrum_pads = ("psd_H1", "psd_L1")
        sink.far_thresh = 1.0
        sink.aggregator_far_thresh = 3.84e-07
        sink.aggregator_far_trials_factor = 1
        sink.output_kafka_server = "localhost:9092"
        sink.gracedb_service_url = None
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = None
        sink.gracedb_cred_reload = True
        sink.gracedb_reload_buffer = 300
        sink.template_sngls = None
        sink.analysis_tag = "test"
        sink.job_type = "noninj"
        sink.analysis_ifos = None
        sink.process_params = None
        sink.delta_t = 0.005
        sink.channel_dict = None
        sink.autocorrelation_lengths = None
        sink.name = "test_sink"

        # Call __post_init__ manually
        gracedb_sink.GraceDBSink.__post_init__(sink)

        mock_producer.assert_called_once()

    @mock.patch("sgnl.sinks.gracedb_sink.GraceDb")
    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.__post_init__")
    def test_init_with_gracedb(self, mock_super_init, mock_gracedb):
        """Test initialization with GraceDB client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.strike_object = mock.MagicMock()
        sink.event_pad = "events"
        sink.spectrum_pads = ("psd_H1", "psd_L1")
        sink.far_thresh = 1.0
        sink.aggregator_far_thresh = 3.84e-07
        sink.aggregator_far_trials_factor = 1
        sink.output_kafka_server = None
        sink.gracedb_service_url = "https://gracedb.ligo.org/api/"
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = None
        sink.gracedb_cred_reload = True
        sink.gracedb_reload_buffer = 300
        sink.template_sngls = None
        sink.analysis_tag = "test"
        sink.job_type = "noninj"
        sink.analysis_ifos = None
        sink.process_params = None
        sink.delta_t = 0.005
        sink.channel_dict = None
        sink.autocorrelation_lengths = None
        sink.name = "test_sink"

        gracedb_sink.GraceDBSink.__post_init__(sink)

        mock_gracedb.assert_called_once()

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.__post_init__")
    def test_init_no_client(self, mock_super_init):
        """Test initialization without any client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.strike_object = mock.MagicMock()
        sink.event_pad = "events"
        sink.spectrum_pads = ("psd_H1", "psd_L1")
        sink.far_thresh = 1.0
        sink.aggregator_far_thresh = 3.84e-07
        sink.aggregator_far_trials_factor = 1
        sink.output_kafka_server = None
        sink.gracedb_service_url = None
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = None
        sink.gracedb_cred_reload = True
        sink.gracedb_reload_buffer = 300
        sink.template_sngls = None
        sink.analysis_tag = "test"
        sink.job_type = "noninj"
        sink.analysis_ifos = None
        sink.process_params = None
        sink.delta_t = 0.005
        sink.channel_dict = None
        sink.autocorrelation_lengths = None
        sink.name = "test_sink"

        gracedb_sink.GraceDBSink.__post_init__(sink)

        assert sink.client is None

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.__post_init__")
    def test_init_assertion_both_clients(self, mock_super_init):
        """Test that assertion fails if both kafka server and gracedb url are truthy."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.strike_object = mock.MagicMock()
        sink.event_pad = "events"
        sink.spectrum_pads = ("psd_H1",)
        sink.far_thresh = 1.0
        sink.aggregator_far_thresh = 3.84e-07
        sink.aggregator_far_trials_factor = 1
        # Both are truthy strings - should fail assertion
        sink.output_kafka_server = "localhost:9092"
        sink.gracedb_service_url = "https://gracedb.ligo.org/api/"
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = None
        sink.gracedb_cred_reload = True
        sink.gracedb_reload_buffer = 300
        sink.template_sngls = None
        sink.analysis_tag = "test"
        sink.job_type = "noninj"
        sink.analysis_ifos = None
        sink.process_params = None
        sink.delta_t = 0.005
        sink.channel_dict = None
        sink.autocorrelation_lengths = None
        sink.name = "test_sink"

        with pytest.raises(AssertionError):
            gracedb_sink.GraceDBSink.__post_init__(sink)

    def test_pull_eos(self):
        """Test pull method with EOS frame."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.event_pad = "events"
        sink.rsnks = {"pad1": "events"}
        sink.mark_eos = mock.MagicMock()
        sink.events = None

        frame = mock.MagicMock()
        frame.EOS = True
        frame.metadata = {"psd": None}

        sink.pull("pad1", frame)

        sink.mark_eos.assert_called_once_with("pad1")
        assert sink.events == frame

    def test_pull_psd(self):
        """Test pull method with PSD frame."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.event_pad = "events"
        sink.rsnks = {"psd_pad": "psd_H1"}
        sink.psds = {}
        sink.mark_eos = mock.MagicMock()

        mock_psd = mock.MagicMock()
        frame = mock.MagicMock()
        frame.EOS = False
        frame.metadata = {"psd": mock_psd}

        sink.pull("psd_pad", frame)

        assert sink.psds["psd_H1"] == mock_psd

    def test_pull_psd_none(self):
        """Test pull method with None PSD in frame."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.event_pad = "events"
        sink.rsnks = {"psd_pad": "psd_H1"}
        sink.psds = {}
        sink.mark_eos = mock.MagicMock()

        frame = mock.MagicMock()
        frame.EOS = False
        frame.metadata = {"psd": None}

        sink.pull("psd_pad", frame)

        # PSD should not be added since it's None
        assert "psd_H1" not in sink.psds

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    @mock.patch("sgnl.sinks.gracedb_sink.event_trigs_to_coinc_xmldoc")
    @mock.patch("sgnl.sinks.gracedb_sink.best_event")
    @mock.patch("sgnl.sinks.gracedb_sink.publish_gracedb")
    @mock.patch("lal.series.make_psd_xmldoc")
    def test_internal_with_gracedb_client(
        self,
        mock_make_psd,
        mock_publish,
        mock_best_event,
        mock_event_to_xmldoc,
        mock_exchange_state,
    ):
        """Test internal method with GraceDB client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": 1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = ""
        sink.events = mock.MagicMock()
        sink.events.events = [
            {
                "event": mock.MagicMock(),
                "trigger": mock.MagicMock(),
                "snr_ts": mock.MagicMock(),
            }
        ]
        sink.public_far_threshold = 1e-6
        sink.sngls_dict = {}
        sink.analysis_ifos = ["H1", "L1"]
        sink.process_params = {}
        sink.delta_t = 0.005
        sink.channel_dict = {"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"}
        sink.autocorrelation_lengths = {}
        sink.psds = {"H1": mock.MagicMock(), "L1": mock.MagicMock()}
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = []
        sink.strike_object = mock.MagicMock()

        mock_event = {"time_subthresh": 1000000000_000000000}
        mock_best_event.return_value = (
            mock_event,
            [{"ifo": "H1"}],
            {"H1": mock.MagicMock()},
        )

        mock_xmldoc = mock.MagicMock()
        mock_event_to_xmldoc.return_value = mock_xmldoc

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            sink.internal()

        mock_publish.assert_called_once()

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    @mock.patch("sgnl.sinks.gracedb_sink.event_trigs_to_coinc_xmldoc")
    @mock.patch("sgnl.sinks.gracedb_sink.best_event")
    @mock.patch("sgnl.sinks.gracedb_sink.publish_kafka")
    @mock.patch("lal.series.make_psd_xmldoc")
    def test_internal_with_kafka_client(
        self,
        mock_make_psd,
        mock_publish,
        mock_best_event,
        mock_event_to_xmldoc,
        mock_exchange_state,
    ):
        """Test internal method with Kafka client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": 1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = "localhost:9092"
        sink.events = mock.MagicMock()
        sink.events.events = [
            {
                "event": mock.MagicMock(),
                "trigger": mock.MagicMock(),
                "snr_ts": mock.MagicMock(),
            }
        ]
        sink.public_far_threshold = 1e-6
        sink.sngls_dict = {}
        sink.analysis_ifos = ["H1", "L1"]
        sink.process_params = {}
        sink.delta_t = 0.005
        sink.channel_dict = {"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"}
        sink.autocorrelation_lengths = {}
        sink.psds = {"H1": mock.MagicMock(), "L1": mock.MagicMock()}
        sink.gracedb_group = "Test"
        sink.gracedb_search = "AllSky"
        sink.job_type = "noninj"
        sink.analysis_tag = "test"
        sink.strike_object = mock.MagicMock()

        mock_event = {"time_subthresh": 1000000000_000000000}
        mock_best_event.return_value = (
            mock_event,
            [{"ifo": "H1"}],
            {"H1": mock.MagicMock()},
        )

        mock_xmldoc = mock.MagicMock()
        mock_event_to_xmldoc.return_value = mock_xmldoc

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            sink.internal()

        mock_publish.assert_called_once()

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    @mock.patch("sgnl.sinks.gracedb_sink.best_event")
    def test_internal_no_event(self, mock_best_event, mock_exchange_state):
        """Test internal method when no event passes threshold."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": 1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = "localhost:9092"
        sink.events = mock.MagicMock()
        sink.events.events = [
            {
                "event": mock.MagicMock(),
                "trigger": mock.MagicMock(),
                "snr_ts": mock.MagicMock(),
            }
        ]
        sink.public_far_threshold = 1e-6

        mock_best_event.return_value = (None, None, None)

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            sink.internal()

        # Should not attempt to publish
        # Just check no errors

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    def test_internal_skips_none_buffer(self, mock_exchange_state):
        """Test internal method skips None buffers and buffers with None event."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": 1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = "localhost:9092"
        sink.events = mock.MagicMock()
        # Include None buffer and buffer with None event
        sink.events.events = [
            None,  # Should be skipped
            {"event": None, "trigger": mock.MagicMock(), "snr_ts": mock.MagicMock()},
        ]
        sink.public_far_threshold = 1e-6

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            # Should not raise any errors - just skip the None buffers
            sink.internal()

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    def test_internal_negative_far_threshold(self, mock_exchange_state):
        """Test internal method with negative FAR threshold."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": -1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = ""

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            sink.internal()

        # Should not attempt to find or publish events

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    def test_internal_flush_kafka_at_eos(self, mock_exchange_state):
        """Test that Kafka client is flushed at EOS."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": -1.0}
        sink.client = mock.MagicMock()
        sink.output_kafka_server = "localhost:9092"

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=True
        ):
            sink.internal()

        sink.client.flush.assert_called_once()

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.exchange_state")
    def test_internal_no_client(self, mock_exchange_state):
        """Test internal method with no client."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        sink.name = "test"
        sink.state = {"far-threshold": 1.0}
        sink.client = None

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock, return_value=False
        ):
            # Should not raise any errors
            sink.internal()


class TestEventTrigsToCoincXmldoc:
    """Tests for event_trigs_to_coinc_xmldoc function."""

    def _create_mock_sngl(self):
        """Helper to create mock sngl with all columns."""

        class MockSngl:
            pass

        mock_sngl = MockSngl()
        for col in lsctables.SnglInspiralTable.validcolumns:
            col_name = col.split(":")[-1]
            setattr(mock_sngl, col_name, 0)
        mock_sngl.mass1 = 1.4
        mock_sngl.mass2 = 1.4
        mock_sngl.mchirp = 1.2
        mock_sngl.Gamma1 = 1.0
        mock_sngl.f_final = 1024.0  # Required for downsample_snr
        return mock_sngl

    def _create_snr_ts(self, epoch_sec=1000000000, length=4097, sample_rate=4096):
        """Helper to create SNR time series."""
        snr_ts = lal.CreateCOMPLEX8TimeSeries(
            name="snr",
            epoch=LIGOTimeGPS(epoch_sec, 0),
            f0=0.0,
            deltaT=1.0 / sample_rate,
            sampleUnits=lal.DimensionlessUnit,
            length=length,
        )
        snr_ts.data.data = numpy.ones(length, dtype=numpy.complex64) * (1 + 0j)
        return snr_ts

    def test_basic_conversion(self):
        """Test basic conversion of event/triggers to XML doc."""
        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}
        snr_ts = self._create_snr_ts()
        snr_ts_dict = {"H1": snr_ts}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None
        # Check that required tables exist
        assert len(lsctables.SnglInspiralTable.get_table(result)) == 1
        assert len(lsctables.CoincInspiralTable.get_table(result)) == 1
        assert len(lsctables.CoincTable.get_table(result)) == 1

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_with_subthreshold_trigger(self, mock_light_travel):
        """Test with subthreshold trigger detection."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data[2048] = 5 + 0j  # Add a peak

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None
        sngl_table = lsctables.SnglInspiralTable.get_table(result)
        assert len(sngl_table) >= 1

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_not_enough_samples(self, mock_light_travel):
        """Test subthreshold search with insufficient samples."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        # Very short L1 series with different epoch - triggers "not enough samples"
        snr_ts_l1 = self._create_snr_ts(epoch_sec=1000000002, length=10)

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_zero_maxsnr(self, mock_light_travel):
        """Test subthreshold search with zero max SNR."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data = numpy.zeros(4097, dtype=numpy.complex64)

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_padding_front(self, mock_light_travel):
        """Test subthreshold search with padding at front."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_010000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_010000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data[10] = 5 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_padding_end(self, mock_light_travel):
        """Test subthreshold search with padding at end."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_990000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_990000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data[4090] = 5 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_not_enough_samples_for_snippet(self, mock_light_travel):
        """Test subthreshold search when peak is too close to edge for bayestar."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_002000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_002000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data[5] = 5 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_exception_handling(self, mock_light_travel):
        """Test that exceptions in subthreshold search are handled."""
        mock_light_travel.side_effect = Exception("Test exception")

        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data = numpy.ones(4097, dtype=numpy.complex64) * (5 + 0j)

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None


class TestTemplateSnglsInit:
    """Tests for template_sngls initialization in GraceDBSink."""

    @mock.patch("sgnl.sinks.gracedb_sink.HTTPControlSinkElement.__post_init__")
    def test_template_sngls_dict_building(self, mock_super_init):
        """Test that template_sngls are correctly parsed into sngls_dict."""
        with mock.patch.object(gracedb_sink.GraceDBSink, "__post_init__"):
            sink = object.__new__(gracedb_sink.GraceDBSink)

        mock_sngl1 = mock.MagicMock()
        mock_sngl2 = mock.MagicMock()

        sink.strike_object = mock.MagicMock()
        sink.event_pad = "events"
        sink.spectrum_pads = ("psd_H1",)
        sink.far_thresh = 1.0
        sink.aggregator_far_thresh = 3.84e-07
        sink.aggregator_far_trials_factor = 1
        sink.output_kafka_server = None
        sink.gracedb_service_url = None
        sink.gracedb_group = "Test"
        sink.gracedb_pipeline = "SGNL"
        sink.gracedb_search = "AllSky"
        sink.gracedb_label = None
        sink.gracedb_cred_reload = True
        sink.gracedb_reload_buffer = 300
        sink.template_sngls = [{1: mock_sngl1}, {2: mock_sngl2}]
        sink.analysis_tag = "test"
        sink.job_type = "noninj"
        sink.analysis_ifos = None
        sink.process_params = None
        sink.delta_t = 0.005
        sink.channel_dict = None
        sink.autocorrelation_lengths = None
        sink.name = "test_sink"

        gracedb_sink.GraceDBSink.__post_init__(sink)

        assert sink.sngls_dict[1] == mock_sngl1
        assert sink.sngls_dict[2] == mock_sngl2


class TestSubthresholdEdgeCases:
    """Tests for edge cases in subthreshold trigger processing."""

    def _create_mock_sngl(self):
        """Helper to create mock sngl with all columns."""

        class MockSngl:
            pass

        mock_sngl = MockSngl()
        for col in lsctables.SnglInspiralTable.validcolumns:
            col_name = col.split(":")[-1]
            setattr(mock_sngl, col_name, 0)
        mock_sngl.mass1 = 1.4
        mock_sngl.mass2 = 1.4
        mock_sngl.mchirp = 1.2
        mock_sngl.Gamma1 = 1.0
        mock_sngl.f_final = 1024.0
        return mock_sngl

    def _create_snr_ts(self, epoch_sec=1000000000, length=4097, sample_rate=4096):
        """Helper to create SNR time series."""
        snr_ts = lal.CreateCOMPLEX8TimeSeries(
            name="snr",
            epoch=LIGOTimeGPS(epoch_sec, 0),
            f0=0.0,
            deltaT=1.0 / sample_rate,
            sampleUnits=lal.DimensionlessUnit,
            length=length,
        )
        snr_ts.data.data = numpy.ones(length, dtype=numpy.complex64) * (1 + 0j)
        return snr_ts

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_multiple_triggerless_ifos(self, mock_light_travel):
        """Test with multiple triggerless IFOs to hit subthresh_trigs iteration."""
        mock_light_travel.return_value = 0.01

        event = {
            "time_subthresh": 1000000000_500000000,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        # Only H1 has a real trigger; L1 and V1 are triggerless
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": 1000000000_500000000,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data[2048] = 5 + 0j  # Peak for L1
        snr_ts_v1 = self._create_snr_ts()
        snr_ts_v1.data.data[2048] = 4 + 0j  # Peak for V1

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1, "V1": snr_ts_v1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1", "V1"],
            process_params={},
            delta_t=0.005,
            channel_dict={
                "H1": "H1:GDS-CALIB_STRAIN",
                "L1": "L1:GDS-CALIB_STRAIN",
                "V1": "V1:Hrec_hoft_16384Hz",
            },
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None
        sngl_table = lsctables.SnglInspiralTable.get_table(result)
        # Should have H1 + L1 + V1 = 3 triggers
        assert len(sngl_table) == 3

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_bayestar_not_enough_samples(self, mock_light_travel):
        """Test when peak is too close to edge for Bayestar (lines 516-522)."""
        mock_light_travel.return_value = 0.01

        # Trigger time close to epoch start so coincidence segment is near beginning
        # epoch=1000000000, trigger at 1000000000.020
        # coinc_segment ~ [0.020-0.015, 0.020+0.015] = [0.005, 0.035] from epoch
        # idx0 ~ 20, idxf ~ 143 (at 4096 Hz)
        # If peak is at idx 50 within this segment, peak_idx = 50 < min_num_samps=109
        trigger_time_ns = 1000000000_020000000
        event = {
            "time_subthresh": trigger_time_ns,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": trigger_time_ns,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        # L1 with peak at idx 50 (within coincidence segment, < min_num_samps)
        snr_ts_l1 = self._create_snr_ts()
        snr_ts_l1.data.data = numpy.zeros(4097, dtype=numpy.complex64)
        snr_ts_l1.data.data[50] = 10 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            autocorrelation_lengths={"0001": 101},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_snr_padding_front(self, mock_light_travel):
        """Test SNR padding at front when snip0 < 0 (lines 527-533)."""
        mock_light_travel.return_value = 0.01

        # Trigger at 0.040s from epoch puts coincidence segment at [0.025, 0.055]
        # idx range ~ [102, 225] at 4096 Hz
        # Peak at idx 150 will be found, which passes min_num_samps check (109)
        # With half_autocorr=200, snip0 = 150-200 = -50 < 0 → hits padding front
        trigger_time_ns = 1000000000_040000000
        event = {
            "time_subthresh": trigger_time_ns,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": trigger_time_ns,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        # Set peak at index 150: >= min_num_samps (109) but < half_autocorr (200)
        snr_ts_l1.data.data = numpy.zeros(4097, dtype=numpy.complex64)
        snr_ts_l1.data.data[150] = 10 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            # Large autocorrelation so half_autocorr_length=200 > peak_idx=150
            autocorrelation_lengths={"0001": 401},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_snr_padding_end(self, mock_light_travel):
        """Test SNR padding at end when snipf > sniplength (lines 540-546)."""
        mock_light_travel.return_value = 0.01

        # Peak at 3950, epoch=1000000000, dt=1/4096
        # Peak time from epoch = 3950/4096 ≈ 0.965s
        # Trigger at 0.965s puts coinc segment at [0.950, 0.980]
        # idx range ~ [3891, 4014] at 4096 Hz, so peak at 3950 is found
        # sniplength=4097, peak_idx=3950, sniplength-peak_idx=147 >= 109 (passes check)
        # With half_autocorr=200, snipf = 3950+200+1 = 4151 > 4097 → padding end
        trigger_time_ns = 1000000000_965000000
        event = {
            "time_subthresh": trigger_time_ns,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": trigger_time_ns,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        # Peak at idx 3950: close to end
        snr_ts_l1.data.data = numpy.zeros(4097, dtype=numpy.complex64)
        snr_ts_l1.data.data[3950] = 10 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        result = gracedb_sink.event_trigs_to_coinc_xmldoc(
            event,
            trigs,
            snr_ts_dict,
            sngls_dict,
            analysis_ifos=["H1", "L1"],
            process_params={},
            delta_t=0.005,
            channel_dict={"H1": "H1:GDS-CALIB_STRAIN", "L1": "L1:GDS-CALIB_STRAIN"},
            # Large autocorrelation so half_autocorr_length=200
            autocorrelation_lengths={"0001": 401},
        )

        assert result is not None

    @mock.patch("sgnl.sinks.gracedb_sink.light_travel_time")
    def test_subthresh_zero_length_snippet_caught(self, mock_light_travel):
        """Test zero length snippet triggers error (lines 556-565).

        The ValueError is caught by except block and function continues.
        """
        mock_light_travel.return_value = 0.01

        # Use trigger time that positions coinc segment for padding front branch
        trigger_time_ns = 1000000000_040000000
        event = {
            "time_subthresh": trigger_time_ns,
            "network_snr_subthresh": 10.0,
            "combined_far": 1e-10,
            "false_alarm_probability": 0.0,
            "likelihood": 100.0,
        }
        trigs = [
            {
                "_filter_id": 1,
                "ifo": "H1",
                "time": trigger_time_ns,
                "phase": 0.0,
                "chisq": 1.0,
                "snr": 8.0,
            }
        ]

        mock_sngl = self._create_mock_sngl()
        sngls_dict = {1: mock_sngl}

        snr_ts_h1 = self._create_snr_ts()
        snr_ts_l1 = self._create_snr_ts()
        # Peak at idx 150 - will trigger padding front with large autocorr
        snr_ts_l1.data.data = numpy.zeros(4097, dtype=numpy.complex64)
        snr_ts_l1.data.data[150] = 10 + 0j

        snr_ts_dict = {"H1": snr_ts_h1, "L1": snr_ts_l1}

        # Mock numpy.pad to return empty array which triggers ValueError
        # The exception is caught by except block (lines 590-591), not raised
        def mock_pad(*args, **kwargs):
            return numpy.array([], dtype=numpy.complex64)

        with mock.patch.object(gracedb_sink.numpy, "pad", side_effect=mock_pad):
            # Exception is caught internally, function should complete
            result = gracedb_sink.event_trigs_to_coinc_xmldoc(
                event,
                trigs,
                snr_ts_dict,
                sngls_dict,
                analysis_ifos=["H1", "L1"],
                process_params={},
                delta_t=0.005,
                channel_dict={
                    "H1": "H1:GDS-CALIB_STRAIN",
                    "L1": "L1:GDS-CALIB_STRAIN",
                },
                # Large autocorrelation to trigger padding branch
                autocorrelation_lengths={"0001": 401},
            )

        assert result is not None



"""Tests for sgnl.dags.config"""

import sys
from unittest import mock

import pytest
from igwn_segments import segment, segmentlist, segmentlistdict

# Mock ezdag before importing config (ezdag may not be installed)
sys.modules["ezdag"] = mock.MagicMock()

from sgnl.dags import config  # noqa: E402


class TestBuildConfig:
    """Tests for build_config function."""

    def test_build_config_basic(self, tmp_path):
        """Test basic config building."""
        # Create a minimal config file
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
condor:
  accounting-group: ligo.dev.o4.test
instruments: H1L1
start: 1000000000
stop: 1000010000
source:
  frame-segments-file: segments.xml.gz
"""
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist(
            [segment(1000000000, 1000005000), segment(1000005000, 1000010000)]
        )
        mock_segments["L1"] = segmentlist([segment(1000000000, 1000010000)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
            mock.patch.object(
                config.segments,
                "split_segments_by_lock",
                return_value=[segment(1000000000, 1000010000)],
            ),
            mock.patch.object(
                config.segments,
                "analysis_segments",
                return_value=segmentlistdict(),
            ),
        ):
            result = config.build_config(
                str(config_file), str(tmp_path), force_segments=True
            )

        assert result.condor.accounting_group == "ligo.dev.o4.test"
        assert result.ifo_list == ["H1", "L1"]
        assert result.span == segment(1000000000, 1000010000)

    def test_build_config_missing_condor_raises(self, tmp_path):
        """Test that missing condor section raises AssertionError."""
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
instruments: H1
"""
        )

        with pytest.raises(AssertionError) as exc_info:
            config.build_config(str(config_file), str(tmp_path))

        assert "condor section" in str(exc_info.value)

    def test_build_config_missing_accounting_group_raises(self, tmp_path):
        """Test that missing accounting-group raises AssertionError."""
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
condor:
  container: /path/to/container
instruments: H1
"""
        )

        with pytest.raises(AssertionError) as exc_info:
            config.build_config(str(config_file), str(tmp_path))

        assert "accounting-group" in str(exc_info.value)

    def test_build_config_no_start_sets_zero_span(self, tmp_path):
        """Test that missing start/stop sets span to (0, 0)."""
        # Create a segments file that will be loaded
        seg_file = tmp_path / "segments.xml.gz"
        seg_file.touch()

        config_file = tmp_path / "config.yml"
        config_file.write_text(
            f"""
condor:
  accounting-group: ligo.dev.o4.test
instruments: H1
source:
  frame-segments-file: {seg_file}
"""
        )

        mock_xmldoc = mock.MagicMock()
        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(0, 100)])

        with (
            mock.patch.object(
                config.ligolw_utils, "load_filename", return_value=mock_xmldoc
            ),
            mock.patch.object(
                config.ligolw_segments,
                "segmenttable_get_by_name",
                return_value=mock_segments,
            ),
        ):
            result = config.build_config(str(config_file), str(tmp_path))

        assert result.span == segment(0, 0)

    def test_build_config_sets_default_paths(self, tmp_path):
        """Test that default paths are set correctly."""
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
condor:
  accounting-group: ligo.dev.o4.test
instruments: H1
source:
  frame-segments-file: segments.xml.gz
"""
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(0, 100)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.build_config(
                str(config_file), str(tmp_path), force_segments=True
            )

        assert result.paths.storage == str(tmp_path)

    def test_build_config_sets_accounting_user(self, tmp_path):
        """Test that accounting_group_user defaults to current user."""
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
condor:
  accounting-group: ligo.dev.o4.test
instruments: H1
source:
  frame-segments-file: segments.xml.gz
"""
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(0, 100)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
            mock.patch.object(config.getpass, "getuser", return_value="testuser"),
        ):
            result = config.build_config(
                str(config_file), str(tmp_path), force_segments=True
            )

        assert result.condor.accounting_group_user == "testuser"

    def test_build_config_min_instruments_default(self, tmp_path):
        """Test that min_instruments_segments defaults to 1."""
        config_file = tmp_path / "config.yml"
        config_file.write_text(
            """
condor:
  accounting-group: ligo.dev.o4.test
instruments: H1
source:
  frame-segments-file: segments.xml.gz
"""
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(0, 100)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.build_config(
                str(config_file), str(tmp_path), force_segments=True
            )

        assert result.min_instruments_segments == 1


class TestCreateSegments:
    """Tests for create_segments function."""

    def test_create_segments_loads_from_file(self, tmp_path):
        """Test loading segments from existing file."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"
        seg_file.touch()

        cfg = DotDict(
            {
                "source": DotDict({"frame_segments_file": str(seg_file)}),
                "span": segment(0, 0),
            }
        )

        mock_xmldoc = mock.MagicMock()
        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(0, 100)])

        with (
            mock.patch.object(
                config.ligolw_utils, "load_filename", return_value=mock_xmldoc
            ),
            mock.patch.object(
                config.ligolw_segments,
                "segmenttable_get_by_name",
                return_value=mock_segments,
            ),
        ):
            result = config.create_segments(cfg)

        assert "H1" in result.segments

    def test_create_segments_force_recreates(self, tmp_path):
        """Test that force_segments recreates even if file exists."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"
        seg_file.touch()

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": None,
                        "use_gwosc_segments": False,
                    }
                ),
                "instruments": ["H1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(0, 0),
            }
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.create_segments(cfg, force_segments=True)

        assert "H1" in result.segments

    def test_create_segments_query_gwosc(self, tmp_path):
        """Test querying GWOSC segments."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": None,
                        "use_gwosc_segments": True,
                    }
                ),
                "instruments": ["H1", "L1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(0, 0),
            }
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])
        mock_segments["L1"] = segmentlist([segment(1000000000, 1000010000)])

        with (
            mock.patch.object(
                config.segments,
                "query_gwosc_segments",
                return_value=mock_segments,
            ) as mock_query,
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.create_segments(cfg)

        mock_query.assert_called_once()
        assert "H1" in result.segments

    def test_create_segments_query_dqsegdb_with_flags(self, tmp_path):
        """Test querying DQSegDB with flags."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": {"H1": "DMT-ANALYSIS_READY:1", "L1": None},
                        "use_gwosc_segments": False,
                    }
                ),
                "instruments": ["H1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(0, 0),
            }
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ) as mock_query,
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.create_segments(cfg)

        mock_query.assert_called_once()
        # Only H1 should have a flag (L1 is None)
        call_args = mock_query.call_args
        assert call_args[0][3] == {"H1": "H1:DMT-ANALYSIS_READY:1"}
        assert "H1" in result.segments

    def test_create_segments_gwosc_with_flags_raises(self, tmp_path):
        """Test that using GWOSC with flags raises ValueError."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": {"H1": "DMT-ANALYSIS_READY:1"},
                        "use_gwosc_segments": True,
                    }
                ),
                "instruments": ["H1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(0, 0),
            }
        )

        with pytest.raises(ValueError) as exc_info:
            config.create_segments(cfg)

        assert "use-gwosc-segments" in str(exc_info.value)

    def test_create_segments_creates_time_bins(self, tmp_path):
        """Test that time bins are created when span is non-zero."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": None,
                        "use_gwosc_segments": False,
                    }
                ),
                "instruments": ["H1"],
                "ifos": ["H1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(1000000000, 1000010000),
                "min_instruments_segments": 1,
            }
        )

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
            mock.patch.object(
                config.segments,
                "split_segments_by_lock",
                return_value=[segment(1000000000, 1000010000)],
            ),
            mock.patch.object(
                config.segments,
                "analysis_segments",
                return_value=segmentlistdict(),
            ),
        ):
            result = config.create_segments(cfg)

        assert hasattr(result, "time_bins")

    def test_create_segments_handles_raw_segments(self, tmp_path):
        """Test that non-segmentlist segments are converted."""
        from sgnl.dags.util import DotDict

        seg_file = tmp_path / "segments.xml.gz"

        cfg = DotDict(
            {
                "source": DotDict(
                    {
                        "frame_segments_file": str(seg_file),
                        "flags": None,
                        "use_gwosc_segments": False,
                    }
                ),
                "instruments": ["H1"],
                "start": 1000000000,
                "stop": 1000010000,
                "span": segment(0, 0),
            }
        )

        # Return raw list of segments instead of segmentlist
        mock_segments = {"H1": [segment(1000000000, 1000010000)]}

        with (
            mock.patch.object(
                config.segments,
                "query_dqsegdb_segments",
                return_value=mock_segments,
            ),
            mock.patch.object(config.segments, "write_segments"),
        ):
            result = config.create_segments(cfg)

        assert isinstance(result.segments["H1"], segmentlist)


class TestApplyCat1Vetoes:
    """Tests for apply_cat1_vetoes function."""

    def test_apply_cat1_vetoes_no_file_specified(self, tmp_path):
        """Test that no vetoes file returns segments unchanged."""
        from sgnl.dags.util import DotDict

        segdict = segmentlistdict()
        segdict["H1"] = segmentlist([segment(0, 1000)])

        cfg = DotDict({"source": DotDict({})})

        result = config.apply_cat1_vetoes(segdict, cfg)

        assert result is segdict

    def test_apply_cat1_vetoes_file_not_found_raises(self, tmp_path):
        """Test that missing vetoes file raises FileNotFoundError."""
        from sgnl.dags.util import DotDict

        segdict = segmentlistdict()
        segdict["H1"] = segmentlist([segment(0, 1000)])

        cfg = DotDict(
            {"source": DotDict({"cat1_vetoes_file": str(tmp_path / "nonexistent.xml")})}
        )

        with pytest.raises(FileNotFoundError) as exc_info:
            config.apply_cat1_vetoes(segdict, cfg)

        assert "not found" in str(exc_info.value)

    def test_apply_cat1_vetoes_subtracts_vetoes(self, tmp_path):
        """Test that vetoes are properly subtracted from segments."""
        from sgnl.dags.util import DotDict

        # Create science segments
        segdict = segmentlistdict()
        segdict["H1"] = segmentlist([segment(0, 1000)])
        segdict["L1"] = segmentlist([segment(0, 1000)])

        # Create veto segments
        veto_dict = segmentlistdict()
        veto_dict["H1"] = segmentlist([segment(200, 400)])
        veto_dict["L1"] = segmentlist([segment(500, 700)])

        # Write vetoes to file
        vetoes_file = tmp_path / "vetoes.xml.gz"
        config.segments.write_segments(veto_dict, output=str(vetoes_file))

        cfg = DotDict({"source": DotDict({"cat1_vetoes_file": str(vetoes_file)})})

        result = config.apply_cat1_vetoes(segdict, cfg)

        # H1: [0, 1000] - [200, 400] = [0, 200] + [400, 1000]
        assert len(result["H1"]) == 2
        assert result["H1"][0] == segment(0, 200)
        assert result["H1"][1] == segment(400, 1000)

        # L1: [0, 1000] - [500, 700] = [0, 500] + [700, 1000]
        assert len(result["L1"]) == 2
        assert result["L1"][0] == segment(0, 500)
        assert result["L1"][1] == segment(700, 1000)

    def test_apply_cat1_vetoes_missing_ifo_in_vetoes(self, tmp_path):
        """Test vetoes with missing IFO are handled correctly."""
        from sgnl.dags.util import DotDict

        # Create science segments for H1 and L1
        segdict = segmentlistdict()
        segdict["H1"] = segmentlist([segment(0, 1000)])
        segdict["L1"] = segmentlist([segment(0, 1000)])

        # Create veto segments only for H1
        veto_dict = segmentlistdict()
        veto_dict["H1"] = segmentlist([segment(200, 400)])

        # Write vetoes to file
        vetoes_file = tmp_path / "vetoes.xml.gz"
        config.segments.write_segments(veto_dict, output=str(vetoes_file))

        cfg = DotDict({"source": DotDict({"cat1_vetoes_file": str(vetoes_file)})})

        result = config.apply_cat1_vetoes(segdict, cfg)

        # H1 should have vetoes applied
        assert len(result["H1"]) == 2

        # L1 should be unchanged (no vetoes for L1)
        assert len(result["L1"]) == 1
        assert result["L1"][0] == segment(0, 1000)


class TestCreateTimeBins:
    """Tests for create_time_bins function."""

    def test_create_time_bins_basic(self):
        """Test basic time bin creation."""
        from sgnl.dags.util import DotDict

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])
        mock_segments["L1"] = segmentlist([segment(1000000000, 1000010000)])

        cfg = DotDict(
            {
                "ifos": ["H1", "L1"],
                "segments": mock_segments,
                "span": segment(1000000000, 1000010000),
            }
        )

        mock_analysis_segs = segmentlistdict()
        mock_analysis_segs[frozenset(["H1", "L1"])] = segmentlist(
            [segment(1000000512, 1000010000)]
        )

        with (
            mock.patch.object(
                config.segments,
                "split_segments_by_lock",
                return_value=[segment(1000000000, 1000010000)],
            ),
            mock.patch.object(
                config.segments,
                "analysis_segments",
                return_value=mock_analysis_segs,
            ),
        ):
            result = config.create_time_bins(cfg)

        assert hasattr(result, "time_bins")
        assert hasattr(result, "time_boundaries")

    def test_create_time_bins_one_ifo_only(self):
        """Test time bin creation with one_ifo_only=True."""
        from sgnl.dags.util import DotDict

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000100000)])

        cfg = DotDict(
            {
                "ifos": ["H1"],
                "segments": mock_segments,
                "span": segment(1000000000, 1000100000),
            }
        )

        mock_split_segs = segmentlist([segment(1000000512, 1000029000)])

        with (
            mock.patch.object(
                config.segments,
                "split_segments_by_lock",
                return_value=[segment(1000000000, 1000100000)],
            ),
            mock.patch.object(
                config.segments, "split_segments", return_value=mock_split_segs
            ),
        ):
            result = config.create_time_bins(cfg, one_ifo_only=True)

        assert hasattr(result, "time_bins")

    def test_create_time_bins_custom_parameters(self):
        """Test time bin creation with custom parameters."""
        from sgnl.dags.util import DotDict

        mock_segments = segmentlistdict()
        mock_segments["H1"] = segmentlist([segment(1000000000, 1000010000)])

        cfg = DotDict(
            {
                "ifos": ["H1"],
                "segments": mock_segments,
                "span": segment(1000000000, 1000010000),
            }
        )

        with (
            mock.patch.object(
                config.segments,
                "split_segments_by_lock",
                return_value=[segment(1000000000, 1000010000)],
            ),
            mock.patch.object(
                config.segments,
                "analysis_segments",
                return_value=segmentlistdict(),
            ) as mock_analysis,
        ):
            config.create_time_bins(
                cfg,
                start_pad=256,
                overlap=256,
                min_instruments=2,
                one_ifo_length=7200,
            )

        # Check that custom params were passed
        call_kwargs = mock_analysis.call_args[1]
        assert call_kwargs["start_pad"] == 256
        assert call_kwargs["overlap"] == 256
        assert call_kwargs["min_instruments"] == 2
        assert call_kwargs["one_ifo_length"] == 7200



"""Tests for sgnl.bin.ll_trigger_counter"""

import json
import sys
from collections import deque
from unittest import mock

import pytest

# Mock ezdag before importing the module
sys.modules["ezdag"] = mock.MagicMock()

from sgnl.bin import ll_trigger_counter  # noqa: E402


class TestXmlString:
    """Tests for xml_string function."""

    def test_xml_string_returns_utf8(self):
        """Test xml_string returns UTF-8 string."""
        mock_rstat = mock.MagicMock()
        mock_rstat.save_fileobj = mock.MagicMock(
            side_effect=lambda f: f.write(b"<xml>test</xml>")
        )

        result = ll_trigger_counter.xml_string(mock_rstat)

        assert result == "<xml>test</xml>"
        mock_rstat.save_fileobj.assert_called_once()


class TestParseCommandLine:
    """Tests for parse_command_line function."""

    def test_default_values(self):
        """Test parsing with default values."""
        with mock.patch("sys.argv", ["ll_trigger_counter"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.topic == "coinc"
            assert args.output_period == 3600.0
            assert args.output == "zerolag_rankingstatpdf.xml.gz"
            assert args.tag == "test"
            assert args.verbose is False

    def test_topic_option(self):
        """Test parsing with --topic option."""
        with mock.patch("sys.argv", ["ll_trigger_counter", "--topic", "custom_topic"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.topic == "custom_topic"

    def test_output_period_option(self):
        """Test parsing with --output-period option."""
        with mock.patch("sys.argv", ["ll_trigger_counter", "--output-period", "1800"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.output_period == 1800.0

    def test_kafka_server_option(self):
        """Test parsing with --kafka-server option."""
        with mock.patch(
            "sys.argv", ["ll_trigger_counter", "--kafka-server", "localhost:9092"]
        ):
            args = ll_trigger_counter.parse_command_line()
            assert args.kafka_server == "localhost:9092"

    def test_output_option(self):
        """Test parsing with --output option."""
        with mock.patch(
            "sys.argv", ["ll_trigger_counter", "--output", "custom_output.xml.gz"]
        ):
            args = ll_trigger_counter.parse_command_line()
            assert args.output == "custom_output.xml.gz"

    def test_tag_option(self):
        """Test parsing with --tag option."""
        with mock.patch("sys.argv", ["ll_trigger_counter", "--tag", "prod"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.tag == "prod"

    def test_verbose_flag(self):
        """Test parsing with -v flag."""
        with mock.patch("sys.argv", ["ll_trigger_counter", "-v"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.verbose is True

    def test_verbose_long_flag(self):
        """Test parsing with --verbose flag."""
        with mock.patch("sys.argv", ["ll_trigger_counter", "--verbose"]):
            args = ll_trigger_counter.parse_command_line()
            assert args.verbose is True


class TestZeroLagCounts:
    """Tests for ZeroLagCounts class."""

    @pytest.fixture
    def mock_dependencies(self):
        """Set up common mocks for ZeroLagCounts tests."""
        with (
            mock.patch("sgnl.bin.ll_trigger_counter.events.EventProcessor.__init__"),
            mock.patch("sgnl.bin.ll_trigger_counter.os.access", return_value=True),
            mock.patch(
                "sgnl.bin.ll_trigger_counter.ligolw_utils.local_path_from_url",
                return_value="/path/to/file",
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.far.RankingStatPDF") as mock_far,
            mock.patch(
                "sgnl.bin.ll_trigger_counter.HTTPControl.exchange_state"
            ) as mock_exchange,
            mock.patch("sgnl.bin.ll_trigger_counter.xml_string", return_value="<xml/>"),
        ):
            mock_rstat = mock.MagicMock()
            mock_far.load.return_value = mock_rstat

            yield {
                "mock_far": mock_far,
                "mock_rstat": mock_rstat,
                "mock_exchange": mock_exchange,
            }

    def test_init_with_defaults(self, mock_dependencies):
        """Test ZeroLagCounts initialization with default parameters."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        assert zlc.process_params == {}
        assert zlc.fname == "zerolag_rankingstatpdf.xml.gz"
        assert zlc.verbose is False
        assert zlc.output_period == 3600
        assert isinstance(zlc.counts_dict, dict)
        assert isinstance(zlc.msgs, deque)

    def test_init_with_process_params(self, mock_dependencies):
        """Test ZeroLagCounts initialization with process params."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()
        process_params = {"key": "value"}

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
            process_params=process_params,
        )

        assert zlc.process_params == {"key": "value"}

    def test_init_file_not_accessible(self):
        """Test ZeroLagCounts raises error when file not accessible."""
        with (
            mock.patch("sgnl.bin.ll_trigger_counter.events.EventProcessor.__init__"),
            mock.patch("sgnl.bin.ll_trigger_counter.os.access", return_value=False),
            mock.patch(
                "sgnl.bin.ll_trigger_counter.ligolw_utils.local_path_from_url",
                return_value="/path/to/file",
            ),
        ):
            mock_logger = mock.MagicMock()
            mock_sgn_control = mock.MagicMock()

            with pytest.raises(ValueError) as exc_info:
                ll_trigger_counter.ZeroLagCounts(
                    kafka_server="localhost:9092",
                    logger=mock_logger,
                    sgn_control=mock_sgn_control,
                )

            assert "zerolag ranking stat PDF does not exist" in str(exc_info.value)

    def test_ingest_message_with_key(self, mock_dependencies):
        """Test ingest method with message that has a key."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        mock_message = mock.MagicMock()
        mock_message.key.return_value = b"key1.key2"
        mock_message.topic.return_value = "test_topic"
        mock_message.partition.return_value = 0
        mock_message.offset.return_value = 100
        mock_message.value.return_value = json.dumps({"data": "value"}).encode("utf-8")
        mock_message.timestamp.return_value = (1234567890, 0)

        zlc.ingest(mock_message)

        assert len(zlc.msgs) == 1
        msg = zlc.msgs[0]
        assert msg.key == ("key1", "key2")
        assert msg.value == {"data": "value"}

    def test_ingest_message_without_key(self, mock_dependencies):
        """Test ingest method with message that has no key."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        mock_message = mock.MagicMock()
        mock_message.key.return_value = None
        mock_message.topic.return_value = "test_topic"
        mock_message.partition.return_value = 0
        mock_message.offset.return_value = 100
        mock_message.value.return_value = json.dumps({"data": "value"}).encode("utf-8")
        mock_message.timestamp.return_value = (1234567890, 0)

        zlc.ingest(mock_message)

        assert len(zlc.msgs) == 1
        msg = zlc.msgs[0]
        assert msg.key == ()

    def test_handle_with_triggers(self, mock_dependencies):
        """Test handle method with triggers."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        # Add some messages
        zlc.msgs.append(mock.MagicMock())

        with mock.patch(
            "sgnl.bin.ll_trigger_counter.aggregator.parse_triggers"
        ) as mock_parse:
            mock_parse.return_value = [{"end": 1000000000.0, "likelihood": 10.5}]

            zlc.handle()

            mock_parse.assert_called_once()
            assert len(zlc.msgs) == 0  # msgs should be cleared

    def test_handle_without_triggers(self, mock_dependencies):
        """Test handle method without triggers."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        with mock.patch(
            "sgnl.bin.ll_trigger_counter.aggregator.parse_triggers"
        ) as mock_parse:
            mock_parse.return_value = []

            zlc.handle()

            mock_logger.debug.assert_called_with("no triggers to process")

    def test_handle_triggers_snapshot(self, mock_dependencies):
        """Test handle method triggers snapshot after output_period."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
            output_period=0,  # Set to 0 to trigger immediate snapshot
        )

        # Set time_since_last to a time in the past
        zlc.time_since_last = 0

        with (
            mock.patch(
                "sgnl.bin.ll_trigger_counter.aggregator.parse_triggers"
            ) as mock_parse,
            mock.patch.object(zlc, "snapshot_output_url") as mock_snapshot,
        ):
            mock_parse.return_value = []

            zlc.handle()

            mock_snapshot.assert_called_once()

    def test_add_coincs_with_likelihood(self, mock_dependencies):
        """Test add_coincs method with coincs that have likelihood."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        coincs = [
            {"end": 1000000005.0, "likelihood": 10.5},
            {"end": 1000000007.0, "likelihood": 15.5},  # Same 10s window, higher LR
            {"end": 1000000015.0, "likelihood": 5.0},  # Different 10s window
        ]

        count = zlc.add_coincs(coincs)

        assert count == 3
        # First two should be in same key (1000000010), max LR should be 15.5
        assert zlc.counts_dict[1000000010.0] == 15.5
        # Third should be in different key
        assert zlc.counts_dict[1000000020.0] == 5.0

    def test_add_coincs_without_likelihood(self, mock_dependencies):
        """Test add_coincs method with coincs that have no likelihood."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        coincs = [
            {"end": 1000000005.0},  # No likelihood
            {"end": 1000000015.0, "likelihood": 5.0},
        ]

        count = zlc.add_coincs(coincs)

        assert count == 1
        assert len(zlc.counts_dict) == 1

    def test_add_counts_to_zerolag_pdf(self, mock_dependencies):
        """Test __add_counts_to_zerolag_pdf method."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        # Add some counts
        zlc.counts_dict = {1000000010.0: 10.5, 1000000020.0: 15.5}

        mock_zlpdf = mock.MagicMock()

        # Call the private method
        zlc._ZeroLagCounts__add_counts_to_zerolag_pdf(mock_zlpdf)

        # Verify count was accessed for each LR value
        assert mock_zlpdf.zero_lag_lr_lnpdf.count.__getitem__.call_count == 2

    def test_snapshot_output_url(self, mock_dependencies):
        """Test snapshot_output_url method."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()
        mock_rstat = mock_dependencies["mock_rstat"]

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        mock_copy = mock.MagicMock()
        mock_rstat.copy.return_value = mock_copy

        with (
            mock.patch("sgnl.bin.ll_trigger_counter.HTTPControl.exchange_state"),
            mock.patch("sgnl.bin.ll_trigger_counter.time_now", return_value=1000000000),
            mock.patch(
                "sgnl.bin.ll_trigger_counter.CacheEntry.from_T050017"
            ) as mock_cache,
            mock.patch("sgnl.bin.ll_trigger_counter.os.path.exists", return_value=True),
            mock.patch(
                "sgnl.bin.ll_trigger_counter.T050017_filename",
                return_value="backup.xml.gz",
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.xml_string", return_value="<xml/>"),
        ):
            mock_entry = mock.MagicMock()
            mock_entry.observatory = "H1"
            mock_entry.description = "test"
            mock_cache.return_value = mock_entry

            zlc.snapshot_output_url()

            mock_rstat.copy.assert_called_once()
            assert mock_copy.save.call_count == 2  # Once for main, once for backup

    def test_snapshot_output_url_creates_backup_dir(self, mock_dependencies):
        """Test snapshot_output_url creates backup directory if not exists."""
        mock_logger = mock.MagicMock()
        mock_sgn_control = mock.MagicMock()
        mock_rstat = mock_dependencies["mock_rstat"]

        zlc = ll_trigger_counter.ZeroLagCounts(
            kafka_server="localhost:9092",
            logger=mock_logger,
            sgn_control=mock_sgn_control,
        )

        mock_copy = mock.MagicMock()
        mock_rstat.copy.return_value = mock_copy

        with (
            mock.patch("sgnl.bin.ll_trigger_counter.HTTPControl.exchange_state"),
            mock.patch("sgnl.bin.ll_trigger_counter.time_now", return_value=1000000000),
            mock.patch(
                "sgnl.bin.ll_trigger_counter.CacheEntry.from_T050017"
            ) as mock_cache,
            mock.patch(
                "sgnl.bin.ll_trigger_counter.os.path.exists", return_value=False
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.os.makedirs") as mock_makedirs,
            mock.patch(
                "sgnl.bin.ll_trigger_counter.T050017_filename",
                return_value="backup.xml.gz",
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.xml_string", return_value="<xml/>"),
        ):
            mock_entry = mock.MagicMock()
            mock_entry.observatory = "H1"
            mock_entry.description = "test"
            mock_cache.return_value = mock_entry

            zlc.snapshot_output_url()

            mock_makedirs.assert_called_once()


class TestMain:
    """Tests for main function."""

    def test_main_setup(self):
        """Test main function setup."""
        with (
            mock.patch(
                "sys.argv",
                [
                    "ll_trigger_counter",
                    "--kafka-server",
                    "localhost:9092",
                    "--output",
                    "test_output.xml.gz",
                ],
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.HTTPControl") as mock_http_control,
            mock.patch("sgnl.bin.ll_trigger_counter.ZeroLagCounts") as mock_zlc_class,
            mock.patch("sgnl.bin.ll_trigger_counter.Queue"),
        ):
            mock_zlc = mock.MagicMock()
            mock_zlc_class.return_value = mock_zlc

            mock_context = mock.MagicMock()
            mock_http_control.return_value.__enter__ = mock.MagicMock(
                return_value=mock_context
            )
            mock_http_control.return_value.__exit__ = mock.MagicMock(return_value=False)

            ll_trigger_counter.main()

            mock_zlc_class.assert_called_once()
            mock_zlc.start.assert_called_once()

    def test_main_verbose(self):
        """Test main function with verbose flag."""
        with (
            mock.patch(
                "sys.argv",
                [
                    "ll_trigger_counter",
                    "--kafka-server",
                    "localhost:9092",
                    "-v",
                ],
            ),
            mock.patch("sgnl.bin.ll_trigger_counter.HTTPControl") as mock_http_control,
            mock.patch("sgnl.bin.ll_trigger_counter.ZeroLagCounts") as mock_zlc_class,
            mock.patch("sgnl.bin.ll_trigger_counter.Queue"),
        ):
            mock_zlc = mock.MagicMock()
            mock_zlc_class.return_value = mock_zlc

            mock_context = mock.MagicMock()
            mock_http_control.return_value.__enter__ = mock.MagicMock(
                return_value=mock_context
            )
            mock_http_control.return_value.__exit__ = mock.MagicMock(return_value=False)

            ll_trigger_counter.main()

            # Verify verbose was passed to ZeroLagCounts
            call_kwargs = mock_zlc_class.call_args[1]
            assert call_kwargs["verbose"] is True



"""Sort and group svd banks and time slices by sample rates."""

# Copyright (C) 2024 Yun-Jing Huang

import itertools
import sys
from collections import Counter, OrderedDict, defaultdict

import numpy as np
import torch
from lal.utils import CacheEntry
from sgnts.base import Offset
from sgnts.transforms.resampler import DOWN_HALF_LENGTH, UP_HALF_LENGTH

from sgnl.svd_bank import Bank, parse_bank_files


def group_and_read_banks(
    svd_bank: list[str],
    source_ifos: list[str] | None = None,
    nsubbank_pretend: int = 0,
    nslice=-1,
    verbose=False,
):
    """Read a list of svd banks file names into bank objects, and group by ifo and
    bankid

    Args:
        svd_bank:
            list[str], a list of svd bank file paths
        source_ifos:
            list[str], a list of ifo the source is processing. Used for checking
            consistency of ifos between the svd banks and data sources
        nsubbank_pretend:
            int, default 0. If not 0, pretend there are this many subbanks by copying
            the first subbank read in this many times. Used for benchmarking.
        nslice:
            int, default -1. If nslice > 0, only read in this number of time slices
        verbose:
            bool, be verbose
    """
    # Read SVD banks
    svd_bank_cache = [CacheEntry.from_T050017(path) for path in svd_bank]
    svd_bank_cache.sort(key=lambda cache_entry: cache_entry.description)
    svd_banks = []

    for _key, seq in itertools.groupby(
        svd_bank_cache, key=lambda cache_entry: cache_entry.description
    ):
        svd_group = dict(
            (cache_entry.observatory, cache_entry.url) for cache_entry in seq
        )
        svd_banks.append(svd_group)

    ifoset = set([tuple(s.keys()) for s in svd_banks])
    if len(ifoset) != 1:
        raise ValueError("The ifos have different sets of svd bank files provided.")

    ifos = sorted(list(ifoset)[0])
    if source_ifos and ifos != sorted(source_ifos):
        raise ValueError(
            f"Data source ifos: {source_ifos} must be the same as svd bank ifos: {ifos}"
        )

    banks: dict[str, list] = {ifo: [] for ifo in ifos}
    if nsubbank_pretend:
        # Pretend we are filtering multiple banks by copying the first bank
        #   many times
        svd_bank_url_dict = svd_banks[0]
        banks_per_svd = parse_bank_files(svd_bank_url_dict, verbose=verbose, fast=True)
        for _ in range(nsubbank_pretend):
            for ifo in ifos:
                banks[ifo].extend([banks_per_svd[ifo][0]])
    else:
        for svd_bank_url_dict in svd_banks:
            banks_per_svd = parse_bank_files(
                svd_bank_url_dict, verbose=verbose, fast=True
            )
            for ifo in ifos:
                banks[ifo].extend(banks_per_svd[ifo])

    # TODO: fix nslice reading
    if nslice > 0:
        for ifo in ifos:
            bifo = banks[ifo]
            for sub in bifo:
                if len(sub.bank_fragments) < nslice:
                    raise ValueError(
                        f"nslice: {nslice} greater than number of bank \
                            fragments: {len(sub.bank_fragments)}"
                    )
                sub.bank_fragments = sub.bank_fragments[:nslice]

    if verbose:
        print("Using nsubbanks", len(banks[ifo]), flush=True, file=sys.stderr)

    return OrderedDict(sorted(banks.items()))


class SortedBank:
    """Sort and group svd banks and time slices by sample rate

    Args:
        banks:
            dict[str, list[Bank]], a dictionary keyed by ifos, with values of lists
            of subbanks as gstlal bank objects
        device:
            str, the torch device
        dtype:
            torch.dtype, the torch data type
        memory_format:
            torch.memory_format, the memory format for the svd tensors
        nsubbank_pretend:
            int, default 0. If not 0, pretend there are this many subbanks by copying
            the first subbank read in this many times. Used for benchmarking.
        nslice:
            int, default -1. If nslice > 0, only read in this number of time slices
        verbose:
            bool, be verbose
    """

    def __init__(
        self,
        banks: dict[str, list[Bank]],
        device: str = "cpu",
        dtype: torch.dtype = torch.float32,
        memory_format: torch.memory_format = torch.contiguous_format,
        nsubbank_pretend: int = 0,
        nslice: int = -1,
        verbose: bool = False,
    ):
        self.device = device
        self.dtype = dtype
        self.memory_format = memory_format
        self.verbose = verbose

        temp = torch.empty(1, dtype=dtype)
        self.cdtype = torch.complex(temp, temp).dtype
        self.nsubbank_pretend = nsubbank_pretend

        self.bank_metadata, self.reordered_bank = self.prepare_metadata(banks)
        # Prepare tensors for LLOID methods
        (
            self.coeff_sv_cat,
            self.bases_cat,
            self.template_ids,
            self.end_time_delta,
            self.subbankids,
            self.bankids_map,
            self.sngls,
            self.autocorrelation_banks,
            self.autocorrelation_length_mask,
            self.autocorrelation_lengths,
            self.horizon_distance_funcs,
            self.template_durations,
        ) = self.prepare_tensors(self.bank_metadata, self.reordered_bank)

        del self.reordered_bank
        del banks

    def prepare_metadata(self, bank):
        """Determine rates and template properties across subbanks

        bank_metadata.keys() = ['ifos', 'nifo', 'nsubbank', 'maxrate', 'unique_rates',
        'nrates', 'nfilter_samples', 'ntempmax', 'delay_per_rate', 'sorted_rates']

        Arguments:
        ----------
        bank:
            gstlal Bank class
        """
        bank_metadata = {}
        ifos = sorted(bank.keys())
        bank_metadata["ifos"] = ifos
        bank_metadata["nifo"] = len(ifos)

        # assume all ifos have same banks
        bank0 = bank[ifos[0]]

        nsubbank = len(bank0)  # number of subbanks in each ifo
        bank_metadata["nsubbank"] = nsubbank

        # determine some properties across banks
        # determine rates
        maxrate = bank0[0].sample_rate_max
        bank_metadata["maxrate"] = maxrate  # assume same for all banks
        unique_rates = dict(
            sorted(
                Counter([bi.rate for b in bank0 for bi in b.bank_fragments]).items(),
                reverse=True,
            )
        )
        bank_metadata["unique_rates"] = unique_rates
        bank_metadata["nrates"] = len(unique_rates)

        # number of samples in each filter, assume it's the same for all templates
        bank_metadata["nfilter_samples"] = (
            bank0[0].bank_fragments[0].orthogonal_template_bank.shape[1]
        )

        # maximum number of templates across all subbanks
        bank_metadata["ntempmax"] = max(
            sub.bank_fragments[0].mix_matrix.shape[1] for sub in bank0
        )
        delay_per_rate = dict.fromkeys(unique_rates.keys())
        for rate in unique_rates:
            delay_per_rate[rate] = max(
                int(bf.start * rate)
                for banki in bank.values()
                for sub in banki
                for bf in sub.bank_fragments
                if rate == bf.rate
            )
        bank_metadata["delay_per_rate"] = delay_per_rate

        sorted_rates, reordered_bank = self.sort_bank_by_rates(bank, bank_metadata)
        bank_metadata["sorted_rates"] = sorted_rates

        return bank_metadata, reordered_bank

    def sort_bank_by_rates(self, bank, bank_metadata):
        """
        Determine the ids to upsample from rate1 to rate2
        This places templates that need to be upsampled from and to
            the same rate next to each other, which allows the templates
            to be upsampled together and allows a simpler slicing method
        The following example shows id placement
        slice  | 0  | 1   |  2 |  3 |
        -----------------------------
        bank0: [2048, 1024, 512, 512]
        bank1: [2048, 1024, 512, 256]
        bank2: [2048,  512, 256]

        bases[sample_rate][upsample_rate]:

        bases[2048][()]
            = [bank0slice0, bank1slice0, bank2slice0]
        bases[1024][(2048,)]
            = [bank0slice1, bank1slice1]
        bases[512][(2048, 1024)]
            = [bank0slice2, bank0slice3, bank1slice2]
            # templates in the same bank will be placed next to each other,
                to allow more convenient summing in sum_same_rates()

        bases[512][(2048,)]
            = [bank2slice1]
        bases[256][(2048, 1024, 512)]
            = [bank1slice3]
        bases[256][(2048, 512)]
            = [bank2slice2]

        * The reason that bases[512][(2048, 1024)] is in a different group than
            bases[512][(2048,)] is because they will go into different upsamplers
        * The reason that bases[256][(2048, 1024, 512)] is in a different group
            than bases[256][(2048, 512)], even though they are upsampling to the
            same rate (512), is because due to the different upsampling path
            256->512->1024->2048 vs. 256->512->2048, the time stamps of the buffers
            will be different (to account for the upsampling padding)

        ----------

        sorted_rates = {
            from_rate: {
                to_rate: {
                    '_ids_counter':,
                    '_addto_counter':,
                    '_addids':,
                    'counts':,
                    'nbmax':,
                    'ntempmax':,
                    'segments':,
                    'sum_same_rate_slices':,
                    'uppad':,
                    'addslice':,
                    'metadata': {
                         bank_order: {
                             'ids':,'nslice':,
                         }
                    },
                }
            }
        }

        """

        ifos = bank_metadata["ifos"]
        bank0 = bank[ifos[0]]
        nsubbank = len(bank0)  # number of subbanks in each ifo
        unique_rates = bank_metadata["unique_rates"]
        maxrate = bank_metadata["maxrate"]

        #
        # Sort the unique rates, this might alter the order of the template banks
        #
        sorted_unique_rates0 = []
        unique_rates_by_bank = []
        for j in range(nsubbank):
            bk = bank0[j]
            if self.nsubbank_pretend:
                bankid = bk.bank_id + "_" + str(j)
            else:
                bankid = bk.bank_id
            unique_rates0 = dict(
                sorted(
                    Counter([bi.rate for bi in bk.bank_fragments]).items(), reverse=True
                )
            )
            unique_rates_by_bank.append(tuple(unique_rates0.keys()))
            # for key in unique_rates:
            #    if key not in unique_rates0:
            #        unique_rates0[key] = 0
            unique_rates0["bankid"] = bankid
            sorted_unique_rates0.append(unique_rates0)

        # sorted_unique_rates = sorted(
        #    # sorted_unique_rates, key=lambda x: x.keys(), reverse=True
        #    sorted_unique_rates,
        #    key=operator.itemgetter(*list(unique_rates.keys())),
        #    reverse=True,
        # )

        unique_rates_by_bank = sorted(set(unique_rates_by_bank), reverse=True)
        sorted_unique_rates = []

        for urb in unique_rates_by_bank:
            for s in sorted_unique_rates0:
                s2 = list(s.keys())
                s2.remove("bankid")
                if list(urb) == s2 and s not in sorted_unique_rates:
                    sorted_unique_rates.append(s)

        # if self.nsubbank_pretend:
        #    sorted_unique_rates *= self.nsubbank_pretend

        if self.verbose:
            print("sorted_unique_rates", flush=True, file=sys.stderr)
            for a in sorted_unique_rates:
                print(a, flush=True, file=sys.stderr)

        # check if nsubbank_pretend is true
        # bankids = [a["bankid"] for a in sorted_unique_rates]
        # nsubbank_pretend = False
        # if len(bankids) > 1 and len(set(bankids)) == 1:
        #    if self.verbose:
        #        print("nsubbank_pretend", flush=True, file=sys.stderr)
        #    nsubbank_pretend = True

        # reorder bankid
        bankid_order = {}
        for i, a in enumerate(sorted_unique_rates):
            bankid = a["bankid"]
            bankid_order[bankid] = i

        reorder = {}
        # reordered bank
        if self.nsubbank_pretend == 0:
            for ifo in ifos:
                reorder[ifo] = {}
                for j in range(nsubbank):
                    bk = bank[ifo][j]
                    bankid = bk.bank_id
                    order = bankid_order[bankid]
                    reorder[ifo][order] = bk
        else:
            if self.verbose:
                print("nsubbank_pretend", flush=True, file=sys.stderr)
            for ifo in ifos:
                reorder[ifo] = {i: b for i, b in enumerate(bank[ifo])}

        #
        # Construct sorted_rates, determine id placement of timeslices
        #
        sorted_rates = defaultdict(dict)
        addto_ids = {}
        for a in sorted_unique_rates:
            temp = {}
            for from_rate in unique_rates:
                from_bank = sorted_rates[from_rate]
                bankid = a["bankid"]
                k = list(a.keys())
                if from_rate in k and a[from_rate] > 0:
                    to_rate = tuple(
                        ki
                        for ki in k
                        if type(ki) is int and ki > from_rate and a[ki] > 0
                    )
                    if to_rate not in from_bank:
                        from_bank[to_rate] = {
                            "_ids_counter": 0,
                            "_addto_counter": 0,
                            "_addids": [],
                            "counts": 0,
                            "nbmax": [],
                            "ntempmax": [],
                            "segments": [],
                            "sum_same_rate_slices": None,
                            "metadata": {},
                        }
                    to_bank = from_bank[to_rate]
                    if from_rate != maxrate:
                        to_bank["counts"] += a[from_rate]
                    temp[from_rate] = to_bank["_addto_counter"]
                    to_bank["_addto_counter"] += 1
            addto_ids[bankid] = temp

        # find the ids each bank should addto after upsampling
        # FIXME: is there a more elegant way?
        for _bankid, a in addto_ids.items():
            for from_rate in unique_rates:
                from_bank = sorted_rates[from_rate]
                k = list(a.keys())
                if from_rate in k:
                    to_rate = tuple(
                        ki for ki in k if type(ki) is int and ki > from_rate
                    )
                    if from_rate != maxrate:
                        from_bank[to_rate]["_addids"].append(a[to_rate[-1]])

        sorted_rates[maxrate][()]["counts"] = nsubbank

        urates = list(unique_rates.keys())
        downpads = {r: None for r in urates}
        downpad = 0
        downpads[urates[0]] = 0
        for urate in urates[1:]:
            downpad += Offset.fromsamples(DOWN_HALF_LENGTH, urate)
            downpads[urate] = downpad

        # loop over all subbanks and update metadata in each rate group
        for j in range(nsubbank):
            bk = reorder[ifos[0]][j]
            rates = [bf.rate for bf in bk.bank_fragments]
            # urates = np.array(sorted(set(rates), reverse=True))
            urates = list(sorted(set(rates), reverse=True))
            urs = urates[1:]
            # uppad = {
            # #r0: sum(Offset.fromsamples(UP_HALF_LENGTH, ri) for ri in urs[urs >= r0])
            # r0: sum([Offset.fromsamples(UP_HALF_LENGTH, ri) for ri in urs[urs >= r0]])
            # for r0 in urs
            # }
            uppads = {r: None for r in urates}
            uppad = 0
            uppads[maxrate] = 0
            for ri in urs:
                uppad += Offset.fromsamples(UP_HALF_LENGTH, ri)
                uppads[ri] = uppad

            for bi, bf in enumerate(bk.bank_fragments):
                rate = bf.rate
                to_rate = tuple(r for r in urates if r > rate)
                rate_group = sorted_rates[rate][to_rate]
                if j not in rate_group["metadata"]:
                    rate_group["metadata"][j] = {
                        "ids": [],
                        "nslice": [],
                    }
                mdata = rate_group["metadata"][j]
                mdata["bankid"] = bk.bank_id
                mdata["nslice"].append(bi)
                rate_group["uppad"] = uppads[rate]
                rate_group["downpad"] = downpads[rate]
                rate_group["shift"] = uppads[rate] + downpads[rate]
                rate_group["segments"].append((bf.start, bf.end))
                if rate != maxrate:
                    if rate == rates[bi - 1]:
                        mdata["ids"].append(rate_group["_ids_counter"])
                        rate_group["_ids_counter"] += 1
                        if len(mdata["ids"]) > 1:
                            rate_group["sum_same_rate_slices"] = []
                    else:
                        mdata["ids"].append(rate_group["_ids_counter"])
                        rate_group["_ids_counter"] += 1
                else:
                    mdata["ids"] = [j]

        for ifo in ifos:
            for j in range(nsubbank):
                bk = reorder[ifo][j]
                rates = [bf.rate for bf in bk.bank_fragments]
                for bf in bk.bank_fragments:
                    rate = bf.rate
                    urates = np.array(sorted(set(rates), reverse=True))
                    to_rate = tuple(r for r in urates if r > rate)
                    rate_group = sorted_rates[rate][to_rate]
                    rate_group["nbmax"].append(bf.mix_matrix.shape[0])
                    rate_group["ntempmax"].append(bf.mix_matrix.shape[1])

        for from_rate, v in sorted_rates.items():
            for _to_rate, rate_group in v.items():
                if from_rate != maxrate:
                    addids = rate_group["_addids"]
                    rate_group["addslice"] = slice(addids[0], addids[-1] + 1)
                rate_group["nbmax"] = max(rate_group["nbmax"])
                rate_group["ntempmax"] = max(rate_group["ntempmax"])
                mdata = rate_group["metadata"]
                if rate_group["sum_same_rate_slices"] is not None:
                    for md in mdata.values():
                        ids = md["ids"]
                        rate_group["sum_same_rate_slices"].append(
                            slice(ids[0], ids[-1] + 1)
                        )

        return sorted_rates, reorder

    def prepare_tensors(self, bank_metadata, reordered_bank):
        """
        Prepare large tensors to store input and output of methods in LLOID
        coeff_sv:      all the coeff_sv from different banks
        bases:         all the bases from different banks
        """
        print(
            "Preparing tensors for LLOID methods...",
            end="",
            flush=True,
            file=sys.stderr,
        )

        dtype = self.dtype
        device = self.device

        ifos = bank_metadata["ifos"]
        nfilter_samples = bank_metadata["nfilter_samples"]
        nsubbank = bank_metadata["nsubbank"]
        sorted_rates = bank_metadata["sorted_rates"]

        # outputs
        bases_by_rate = {
            r1: {r2: {} for r2 in rb.keys()} for r1, rb in sorted_rates.items()
        }
        coeff_sv_by_rate = {
            r1: {r2: {} for r2 in rb.keys()} for r1, rb in sorted_rates.items()
        }

        # construct big tensors of data, bases, and coeff_sv, grouped by sample rates
        for from_rate, rbr in sorted_rates.items():
            for to_rate, rate_group in rbr.items():
                count = rate_group["counts"]
                mdata = rate_group["metadata"]
                nbm = rate_group["nbmax"]
                ntempmax = rate_group["ntempmax"]

                for ifo in ifos:

                    # group the bases by sample rate
                    bases_by_rate[from_rate][to_rate][ifo] = torch.zeros(
                        size=(count, nbm, nfilter_samples),
                        device=device,
                        dtype=dtype,
                    )

                    # group the coeff by sample rate
                    coeff_sv_by_rate[from_rate][to_rate][ifo] = torch.zeros(
                        size=(count, ntempmax, nbm),
                        device=device,
                        dtype=dtype,
                    )

                    # fill in the bases and coeff tensors!
                    # for k, ifo in enumerate(ifos):
                    bifo = reordered_bank[ifo]
                    for bank_order, md in mdata.items():
                        bifo_order = bifo[bank_order]
                        ids = md["ids"]
                        nslices = md["nslice"]
                        for id0, nslice in zip(ids, nslices):
                            this_slice = bifo_order.bank_fragments[nslice]
                            assert this_slice.rate == from_rate
                            b = this_slice.orthogonal_template_bank
                            c = this_slice.mix_matrix.T
                            bases_by_rate[from_rate][to_rate][ifo][
                                id0, : b.shape[0], :
                            ] = torch.tensor(b, device=device, dtype=dtype)
                            coeff_sv_by_rate[from_rate][to_rate][ifo][
                                id0, : c.shape[0], : c.shape[1]
                            ] = torch.tensor(c, device=device, dtype=dtype)
                bases = bases_by_rate[from_rate][to_rate][ifo].view(
                    -1,
                    nfilter_samples,
                )
                bases = (
                    bases.unsqueeze(1).unsqueeze(0).to(memory_format=self.memory_format)
                )
                bases_by_rate[from_rate][to_rate][ifo] = bases.view(
                    count, nbm, nfilter_samples
                )
                # benchmark conv methods
                rate_group["conv_group"] = True
                """
                if self.device == "cpu":
                    rb["conv_group"] = True
                else:
                    if self.verbose:
                        print("Benchmarking conv...", flush=True, file=sys.stderr)
                    tgroup = benchmark.Timer(
                        stmt="SNRSlices.conv_group(mask, data, nbm, basesr)",
                        setup="from greg.filtering.snr_slices import SNRSlices",
                        globals={
                            "mask": [True] * nifo,
                            "data": data_by_rate[from_rate][to_rate],
                            "nbm": nbm,
                            "basesr": bases_by_rate[from_rate][to_rate],
                        },
                    )

                    tgroupm = tgroup.timeit(500).median

                    tloop = benchmark.Timer(
                        stmt="SNRSlices.conv_loop(mask, data, nbm, basesr)",
                        setup="from greg.filtering.snr_slices import SNRSlices",
                        globals={
                            "mask": [True] * nifo,
                            "data": data_by_rate[from_rate][to_rate],
                            "nbm": nbm,
                            "basesr": bases_by_rate[from_rate][to_rate],
                        },
                    )

                    tloopm = tloop.timeit(500).median

                    if self.verbose:
                        print(
                            f"{from_rate=} {to_rate=}",
                            "conv_group",
                            tgroupm,
                            "conv_loop",
                            tloopm,
                            flush=True,
                            file=sys.stderr,
                        )
                    if tgroupm < tloopm:
                        rb["conv_group"] = True
                    else:
                        rb["conv_group"] = False
                """

        # Get template ids
        #   Assume same ids for all ifos for the same bank
        #   Init template ids as -1 for banks with ntemp < ntempmax,
        #   the template id for empty entries will be -1
        ntempmax = bank_metadata["ntempmax"]
        template_ids = (
            torch.ones(size=(nsubbank, ntempmax // 2), dtype=torch.int32) * -1
        )
        subbankids = []
        sngls = []
        end_time_delta = torch.zeros(size=(nsubbank,), dtype=torch.long)
        # in seconds
        template_durations = np.zeros((nsubbank, ntempmax // 2))
        bankids_map = defaultdict(list)
        horizon_distance_funcs = {}
        autocorrelation_lengths = {}
        for j in range(nsubbank):
            sngl = reordered_bank[ifos[0]][j].sngl_inspiral_table
            template_ids0 = torch.tensor([row.template_id for row in sngl])
            template_ids[j, : template_ids0.shape[0]] = template_ids0

            ends0 = [row.end.ns() for row in sngl]
            assert len(set(ends0)) == 1, "there are different end times in a subbank"
            end_time_delta[j] = list(set(ends0))[0]

            subbank_id = reordered_bank[ifos[0]][j].bank_id
            if self.nsubbank_pretend:
                bank_id = "%04d" % int(subbank_id.split("_")[0]) + "_" + str(j // 2)
            else:
                bank_id = "%04d" % int(subbank_id.split("_")[0])
            horizon_distance_funcs[bank_id] = reordered_bank[ifos[0]][
                j
            ].horizon_distance_func
            subbankids.append(subbank_id)
            bankids_map[bank_id].append(j)
            sngl0 = {row.template_id: row for row in sngl}
            sngls.append(sngl0)
            for k, row in enumerate(sngl):
                template_durations[j, k] = row.template_duration
            autocorrelation_lengths[bank_id] = reordered_bank[ifos[0]][
                j
            ].autocorrelation_bank.shape[1]

        bankids_map = OrderedDict(sorted(bankids_map.items()))

        # Write out single inspiral table
        # sngl0 = reordered_bank[ifos[0]][0].sngl_inspiral_table
        # row = sngl0[0]
        # keys = [a for a in dir(row) if not a.startswith('__') and not
        #           callable(getattr(row, a))]
        # import h5py
        # with h5py.File('sngl_inspiral_table.h5', "a") as f:
        #    for j in range(nsubbank):
        #        sngl = reordered_bank[ifos[0]][j].sngl_inspiral_table
        #        for row in sngl:
        #            group = str(row.template_id)
        #            f.create_group(group)
        #            for k in keys:
        #                v = getattr(row, k)
        #                if k == 'end':
        #                    v = float(v)
        #                f[group][k] = v

        # Trigger generator
        # Get the autocorrelation_bank
        acl = [
            reordered_bank[ifo][j].autocorrelation_bank.shape[1]
            for i, ifo in enumerate(ifos)
            for j in range(nsubbank)
        ]
        max_acl = max(acl)
        if len(set(acl)) > 1:
            print("Warning: different autocorrelation lengths among banks")
            autocorrelation_length_mask = {}
        else:
            autocorrelation_length_mask = None

        autocorrelation_banks = {}
        for ifo in ifos:
            autocorrelation_banks[ifo] = torch.zeros(
                size=(nsubbank, ntempmax // 2, max_acl),
                device=device,
                dtype=self.cdtype,
            )
            if autocorrelation_length_mask is not None:
                autocorrelation_length_mask[ifo] = torch.ones(
                    size=(nsubbank, ntempmax // 2, max_acl),
                    device=device,
                    dtype=bool,
                )

        for ifo in ifos:
            for j in range(nsubbank):
                acorr = reordered_bank[ifo][j].autocorrelation_bank
                # this is for adjusting to the bank used for impulse test
                # if acorr.shape[0] > ntempmax // 2:
                #    acorr = acorr[: ntempmax // 2]
                acorr_len = acorr.shape[1]
                if acorr_len < max_acl:
                    pad = (max_acl - acorr_len) // 2
                    autocorrelation_banks[ifo][j, : acorr.shape[0], pad:-pad] = (
                        torch.tensor(acorr)
                    )
                    autocorrelation_length_mask[ifo][j, :, :pad] = False
                    autocorrelation_length_mask[ifo][j, :, -pad:] = False
                else:
                    autocorrelation_banks[ifo][j, : acorr.shape[0], :] = torch.tensor(
                        acorr
                    )
                # autocorrelation_banks[ifo][j, : acorr.shape[0], : acorr.shape[1]] = (
                #     torch.tensor(acorr)
                # )
        print(" Done.", flush=True, file=sys.stderr)

        return (
            coeff_sv_by_rate,
            bases_by_rate,
            template_ids,
            end_time_delta,
            subbankids,
            bankids_map,
            sngls,
            autocorrelation_banks,
            autocorrelation_length_mask,
            autocorrelation_lengths,
            horizon_distance_funcs,
            template_durations,
        )
